import configparser, os, re, pathlib, yaml, cctk, h5py, logging
import numpy as np
import presto

logger = logging.getLogger(__name__)

#### OVERALL PRESTO CONFIGURATION, RUNS ON STARTUP

MAX_QC_ATTEMPTS = 50
QC_TOL = 0.05

# this module loads the configuration file
CONFIGURATION_FILE = "presto.config"
if not os.path.isfile(CONFIGURATION_FILE):
    if "PRESTO_CONFIG" in os.environ:
        CONFIGURATION_FILE = os.environ["PRESTO_CONFIG"]
    else:
        raise ValueError(f"Fatal error: presto configuration file '{CONFIGURATION_FILE}' not found in current directory and $PRESTO_CONFIG not found")

config = configparser.ConfigParser()
config.read(CONFIGURATION_FILE)

# auto-populate some directory names
PRESTO_STARTUP_DIRECTORY = os.getcwd()
USER_HOME_DIRECTORY = str(pathlib.Path.home())
def resolve_directory(directory):
    directory = re.sub("@presto", PRESTO_STARTUP_DIRECTORY, directory)
    directory = re.sub("~", USER_HOME_DIRECTORY, directory)
    return directory

def check_directory(field_name, directory):
    """ Checks if a directory exists."""
    directory_exists = os.path.isdir(directory)
    if not directory_exists:
        logger.error(f"Error in configuration entry for {field_name}: directory {directory} does not exist.")

XTB_PATH = config['xtb']['XTB_PATH']
if XTB_PATH.lower() == "@auto":
    CONDA_PATH = os.getenv('CONDA_PREFIX')
    assert CONDA_PATH is not None, "Could not get a value for CONDA_PATH when trying to autodetect XTBPATH. Aborting. (Do you have the environment activated?)"
    XTB_PATH = f"{CONDA_PATH}/share/xtb"
else:
    XTB_PATH = resolve_directory(config['xtb']['XTB_PATH'])
check_directory("XTB_PATH",XTB_PATH)

# load execs - will be run during Calculator.__init__()
XTB_EXEC = config['xtb']['XTB_EXEC']
G16_EXEC = config['gaussian']['GAUSSIAN_EXEC']
logger.info(f"Loaded configuration data from {CONFIGURATION_FILE}.")

#### JOB-SPECIFIC CONFIGURATION

def build(file, checkpoint, geometry=None, oldchk=None, oldchk_idx=-1, **args):
    """
    Build a *presto* trajectory from a ``.yml`` config file.

    The config file does not specify the actual structure (geometry and atomic numbers). This can come from one of two sources:
    1. An existing ``presto`` trajectory saved as an ``.chk`` file.
    2. An ``.xyz`` file.

    Args:
        file (str): path to config ``yaml`` file
        checkpoint (str): path to checkpoint file for this run
        geometry (str): path to geometry ``.xyz`` file
        oldchk (str): path to checkpoint file for previous run
        oldchk_idx (int): which frame from ``oldchk`` to use as starting point
    """
    assert isinstance(file, str), "``file`` must be a string."
    assert isinstance(checkpoint, str), "``checkpoint`` must be a string."

    settings = dict()
    with open(file, "r+") as f:
        settings = yaml.safe_load(f)

    assert "timestep" in settings, "Need `timestep` in config YAML file."
    assert isinstance(settings["timestep"], (float, int)), "`timestep` must be numeric."
    args["timestep"] = settings["timestep"]

    assert "integrator" in settings, "Need `integrator` in config YAML file."
    assert "calculator" in settings, "Need `calculator` in config YAML file."
    assert "type" in settings, "Need `type` in config YAML file."

    if "active_atoms" in settings:
        args["active_atoms"] = parse_atom_list(settings["active_atoms"])
    elif "inactive_atoms" in settings:
        args["inactive_atoms"] = parse_atom_list(settings["inactive_atoms"])
    else:
        args["inactive_atoms"] = list()

    if "high_atoms" in settings:
        args["high_atoms"] = parse_atom_list(settings["high_atoms"])
    else:
        args["high_atoms"] = None

    assert "stop_time" in settings, "Need `stop_time` in config YAML file."
    assert isinstance(settings["stop_time"], int), "`stop_time` must be an integer."
    args["stop_time"] = settings["stop_time"]

    if "save_interval" in settings:
        assert isinstance(settings["save_interval"], int) and settings["save_interval"] > 0, "`save_interval` must be positive integer"
        args["save_interval"] = settings["save_interval"]

    if "forwards" in settings:
        args["forwards"] = settings["forwards"]

    p = None
    if "potential" in settings:
        p = presto.potentials.build_potential(settings["potential"])

    constraints = list()
    if "constraints" in settings:
        constraints = presto.constraints.build_constraints(settings["constraints"])

    if "anchor" in settings:
        assert isinstance(settings["anchor"], int), "``anchor`` must be an atomic number!"
        constraints.insert(0, presto.constraints.Anchor(settings["anchor"]))

    i = presto.integrators.build_integrator(settings["integrator"], potential=p)
    c = presto.calculators.build_calculator(settings["calculator"], checkpoint, constraints=constraints)

    args["checkpoint_filename"] = checkpoint
    assert isinstance(settings["type"], str), "`type` must be a string"
    if settings["type"].lower() == "equilibration":
        mol, x, v, a = None, None, None, None
        if not os.path.exists(checkpoint):
            if geometry is not None:
                assert isinstance(geometry, str), "``geometry``must be a string!"
                assert re.search("xyz$", geometry), "``geometry`` must be a path to an ``.xyz`` file!"
                mol = cctk.XYZFile.read_file(geometry).molecule
                args["atomic_numbers"] = mol.atomic_numbers
                x = mol.geometry

            elif oldchk is not None:
                assert os.path.exists(oldchk), f"Can't locate checkpoint file at {oldchk}!"
                with h5py.File(oldchk, "r") as h5:
                    args["atomic_numbers"] = h5.attrs["atomic_numbers"].view(cctk.OneIndexedArray)
                    x = h5.get("all_positions")[oldchk_idx].view(cctk.OneIndexedArray)
                    v = h5.get("all_velocities")[oldchk_idx].view(cctk.OneIndexedArray)
                    a = h5.get("all_accelerations")[oldchk_idx].view(cctk.OneIndexedArray)
            else:
                raise ValueError("No checkpoint file; need a geometry ``.xyz`` file or old checkpoint file for this to work!")

        assert "bath_scheduler" in settings, "Need `bath_scheduler` in config YAML file for type=`equilibration`!"
        s = build_bath_scheduler(settings["bath_scheduler"])

        t = presto.trajectory.EquilibrationTrajectory(
            calculator=c,
            integrator=i,
            bath_scheduler=s,
            **args,
        )

        if len(t.frames) == 0:
            t.initialize(x, velocities=v, accelerations=a)

        return t

    if settings["type"].lower() == "reaction":
        if not os.path.exists(checkpoint):
            x, v, a, temp = None, None, None, None
            if "quasiclassical" in settings:
                assert "output_file" in settings["quasiclassical"], f"Need Gaussian output file for quasiclassical initialization!"
                assert os.path.exists(settings["quasiclassical"]["output_file"]), f"Need Gaussian output file for quasiclassical initialization!"

                assert "temperature" in settings["quasiclassical"], "Need temperature for quasiclassical initialization"
                temp = settings["quasiclassical"]["temperature"]
                assert isinstance(temp, (int, float)), "Temperature must be numeric!"

                # thermal excitations using quantum harmonic oscillator. see cctk/quasiclassical.py for details
                for idx in range(MAX_QC_ATTEMPTS):
                    qcf = cctk.GaussianFile.read_file(settings["quasiclassical"]["output_file"])
                    mol = qcf.get_molecule()
                    excited, expected_PE, _, velocities = cctk.quasiclassical.get_quasiclassical_perturbation(mol, return_velocities=True)

                    # check that we're close to expected additional PE, and that our perturbation hasn't done anything wild.
                    # if so we just try again.
                    actual_PE, _ = c.evaluate(mol.atomic_numbers, excited.geometry, args["high_atoms"])
                    extra_PE = (actual_PE - qcf.ensemble[-1, "energy"]) * presto.constants.KCAL_PER_HARTREE
                    diff = abs(expected_PE - extra_PE)
                    if (diff < (QC_TOL * expected_PE)):
                        args["atomic_numbers"] = mol.atomic_numbers
                        x = excited.geometry
                        v = velocities
                        a = np.zeros_like(v.view(np.ndarray)).view(cctk.OneIndexedArray)
                        break
                    else:
                        logger.error(f"Error initializing trajectory -- attempt {idx}/{MAX_QC_ATTEMPTS}")

                    if idx == MAX_QC_ATTEMPTS - 1:
                        logger.error("Could not initialize!")
                        return None

            else:
                assert os.path.exists(oldchk), f"Need old checkpoint file for reaction trajectory!"
                with h5py.File(oldchk, "r") as h5:
                    atomic_numbers = h5.attrs["atomic_numbers"]
                    x = h5.get("all_positions")[oldchk_idx].view(cctk.OneIndexedArray)
                    v = h5.get("all_velocities")[oldchk_idx].view(cctk.OneIndexedArray)
                    a = h5.get("all_accelerations")[oldchk_idx].view(cctk.OneIndexedArray)
                    temp = h5.get("bath_temperatures")[oldchk_idx]

                    args["atomic_numbers"] = atomic_numbers.view(cctk.OneIndexedArray)

        assert "termination_function" in settings, "Need `termination_function` in config YAML file for type=`reaction`!"
        f = build_termination_function(settings["termination_function"])

        if "time_after_finished" in settings:
            assert isinstance(settings["time_after_finished"], (float, int)), "Need `time_after_finished` to be numeric!"
            args["time_after_finished"] = settings["time_after_finished"]

        t = presto.trajectory.ReactionTrajectory(
            calculator=c,
            integrator=i,
            termination_function=f,
            **args,
        )

        if not os.path.exists(checkpoint):
            t.initialize(positions=x, velocities=v, accelerations=a, bath_temp=temp)

        assert len(t.frames) > 0, "reaction trajectory needs a frame to start from!"
        return t

def build_bath_scheduler(settings):
    assert isinstance(settings, dict), "Need to pass a dictionary!!"
    assert "type" in settings, "Need `type` for bath scheduler"
    assert isinstance(settings["type"], str), "Bath scheduler `type` must be a string"

    if settings["type"].lower() == "linear":
        assert "start_temp" in settings, "Need `start_temp` for this bath scheduler!"
        assert isinstance(settings["start_temp"], (float, int)), "`start_temp` must be numeric!"
        assert settings["start_temp"] > 0, "`start_temp` must be positive!"

        assert "target_temp" in settings, "Need `target_temp` for this bath scheduler!"
        assert isinstance(settings["target_temp"], (float, int)), "`target_temp` must be numeric!"
        assert settings["target_temp"] > 0, "`target_temp` must be positive!"

        assert "target_time" in settings, "Need `target_time` for this bath scheduler!"
        assert isinstance(settings["target_time"], (float, int)), "`target_time` must be numeric!"
        assert settings["target_time"] > 0, "`target_time` must be positive!"

        delta = settings["start_temp"] - settings["target_temp"]

        def sched(time):
            if time > settings["target_time"]:
                return settings["target_temp"]
            else:
                return settings["target_temp"] + delta * (1 - time / settings["target_time"])

        return sched

    elif settings["type"].lower() == "constant":
        assert "target_temp" in settings, "Need `target_temp` for this bath scheduler!"
        assert isinstance(settings["target_temp"], (float, int)), "`target_temp` must be numeric!"
        assert settings["target_temp"] > 0, "`target_temp` must be positive!"

        def sched(time):
            return settings["target_temp"]

        return sched

    else:
        raise ValueError(f"Unknown bath scheduler type {settings['type']}! Allowed options are `linear` or `constant`.")

def parse_atom_list(string):
    atoms = []
    for item in string.split(", "):
        try:
            if re.search("-", item):
                start, stop = item.split("-")
                atoms = atoms + list(range(int(start), int(stop) + 1))
            else:
                atoms.append(int(item))
        except Exception as e:
            raise ValueError(f"Error parsing list item {item} - needs to be integers (3) or ranges (4-7).")
    return np.array(list(set(atoms))) # keep only distinct atoms

def build_termination_function(settings):
    def term(frame):
        m = frame.molecule().assign_connectivity()
        for name, row in settings.items():
            exit_code = 1
            if re.search("r$", name):
                exit_code = 2

            words = list(filter(None, row.split(" ")))
            if words[0] == "bond":
                assert len(words) == 3, f"Termination condition ``bond`` must be of form ``bond atom1 atom2`` -- {row} is invalid!"
                if m.get_bond_order(int(words[1]), int(words[2])):
                    return exit_code
            elif words[0] == "distance":
                assert len(words) == 5, f"Termination condition ``bond`` must be of form ``bond atom1 atom2 [greater_than/less_than] value`` -- {row} is invalid!"
                if words[3] == "greater_than":
                    if m.get_distance(int(words[1]), int(words[2])) > float(words[4]):
                        return exit_code
                elif words[3] == "less_than":
                    if m.get_distance(int(words[1]), int(words[2])) < float(words[4]):
                        return exit_code
                else:
                    raise ValueError(f"Invalid operator {words[3]} -- must be ``greater_than`` or ``less_than``")
            else:
                raise ValueError(f"Invalid constraint type {words[0]} -- must be ``bond`` or ``distance``")
        return False

    return term




import numpy as np
import os, random, string, re, cctk, ctypes, copy, shutil, time, tempfile, logging
import subprocess as sp
import multiprocessing as mp
import presto

logger = logging.getLogger(__name__)

############################
### Directory Parameters ###
############################

# A-Za-z0-9
LETTERS_AND_DIGITS = string.ascii_letters + string.digits

# where the xtb GFNn config files are stored
XTB_PATH = presto.config.XTB_PATH

###########################

class Calculator():
    """
    Generic calculator class.
    """

    def __init__(self):
        pass

    def evaluate(self, atomic_numbers, positions, high_atoms, pipe):
        """
        Return energy, forces
        """
        pass

class NullCalculator(Calculator):
    """
    Does nothing except evaluate constraints. Mostly for testing purposes, but I guess you could also model something insanely boring this way.
    """
    def __init__(self, constraints=list()):
        for c in constraints:
            assert isinstance(c, presto.constraints.Constraint), "{c} is not a valid constraint!"
        self.constraints = constraints

    def evaluate(self, atomic_numbers, positions, high_atoms=None, pipe=None):
        energy = 0
        forces = np.zeros_like(positions).view(cctk.OneIndexedArray)

        for c in self.constraints:
            f, e = c.evaluate(positions)
            energy += e
            forces += f

        if pipe is not None:
            assert isinstance(pipe, mp.connection.Connection), "not a valid Connection instance!"
            pipe.send([energy, forces])
            pipe.close()
        else:
            return energy, forces


class XTBCalculator(Calculator):
    """
    Attributes:
        charge (int):
        multiplicity (int):
        gfn (int or str):
        parallel (int):
        xcontrol_path (str):
        topology (str): path to gfn-ff topology file, if gfn is ``ff``.
    """

    def __init__(self, charge=0, multiplicity=1, gfn=2, parallel=1, constraints=list(), xcontrol_path=None, topology=None):
        assert isinstance(charge, int)
        assert isinstance(multiplicity, int)
        assert isinstance(gfn, (int, str))
        assert isinstance(parallel, int)

        self.charge = charge
        self.multiplicity = multiplicity
        self.gfn = gfn
        self.parallel = parallel

        for c in constraints:
            assert isinstance(c, presto.constraints.Constraint), "{c} is not a valid constraint!"
        self.constraints = constraints

        if xcontrol_path is not None:
            assert isinstance(xcontrol_path, str)
        self.xcontrol_path = xcontrol_path

        if gfn == "ff":
            assert isinstance(topology, str), "need path for topology file!"
        self.topology = topology

    def evaluate(self, atomic_numbers, positions, high_atoms=None, pipe=None):
        """
        Gets the electronic energy and cartesian forces for the specified geometry.

        Args:
            atomic_numbers (cctk.OneIndexedArray): the atomic numbers (int)
            positions (cctk.OneIndexedArray): the atomic positions in angstroms
            high_atoms (np.ndarray): do nothing with this
            pipe (): for multiprocessing, the connection through which objects should be returned to the parent process

        Returns:
            energy (float): in Hartree
            forces (cctk.OneIndexedArray): in amu Å per fs**2
            time (float): in seconds
        """
        assert shutil.which(presto.config.XTB_EXEC), f"bad xtb executable {presto.config.XTB_EXEC}"

        old_working_directory = os.getcwd()

        # build xtb command
        command = presto.config.XTB_EXEC
        if self.gfn == "ff":
            command += " --gfnff"
        else:
            command += f" --gfn {self.gfn}"
        command += f" --chrg {self.charge} --uhf {self.multiplicity - 1}"
        if self.parallel > 1:
            command += f" --parallel {self.parallel}"
        if self.xcontrol_path:
            command += f" --input {self.xcontrol_path}"
        command += " --grad xtb-in.xyz &> xtb-out.out"

        # set env variables
        os.environ["OMP_NUM_THREADS"] = str(self.parallel)
        os.environ["MKL_NUM_THREADS"] = str(self.parallel)
        os.environ["OMP_STACKSIZE"] = "4G"
        os.environ["OMP_MAX_ACTIVE_LEVELS"] = "1"
        os.environ["XTBPATH"] = XTB_PATH
        os.environ["XTBHOME"] = XTB_PATH

        energy, forces = None, None
        elapsed = 0
        with tempfile.TemporaryDirectory() as tmpdir:
            if self.topology and os.path.exists(self.topology):
                shutil.copyfile(self.topology, f"{tmpdir}/gfnff_topo")

            molecule = cctk.Molecule(atomic_numbers, positions, charge=self.charge, multiplicity=self.multiplicity)
            cctk.XYZFile.write_molecule_to_file(f"{tmpdir}/xtb-in.xyz", molecule)

            # run xtb
            try:
                start = time.time()
                result = sp.run(command, cwd=tmpdir, shell=True, capture_output=True)
                end = time.time()
                elapsed = end - start

                result.check_returncode()
                assert os.path.isfile(f"{tmpdir}/energy"), "no energy file!"
                assert os.path.isfile(f"{tmpdir}/gradient"), "no gradient file!"
            except Exception as e:
                shutil.copyfile(f"{tmpdir}/xtb-in.xyz", f"{old_working_directory}/xtb-failed-input.xyz")
                shutil.copyfile(f"{tmpdir}/xtb-out.out", f"{old_working_directory}/xtb-failed-output.out")
                raise ValueError(f"xtb failed:\n{e}\nfiles:{os.listdir(tmpdir)}")

            # parse energy
            with open(f"{tmpdir}/energy", "r") as f:
                energy_lines = f.read().splitlines()
            energy = float(energy_lines[1].split()[1])

            # parse forces
            with open(f"{tmpdir}/gradient", "r") as f:
                gradient_lines = f.read().splitlines()
            forces = []
            for line in gradient_lines:
                fields = line.split()
                if len(fields) == 3:
                    x,y,z = float(fields[0]), float(fields[1]), float(fields[2])
                    forces.append([-x,-y,-z])
            forces = cctk.OneIndexedArray(forces)
            forces = forces * presto.constants.AMU_A2_FS2_PER_HARTREE_BOHR
            assert len(forces) == molecule.get_n_atoms(), "unexpected number of atoms"

            if self.gfn == "ff" and not os.path.exists(self.topology):
                assert os.path.exists(f"{tmpdir}/gfnff_topo"), "xtb didn't generate topology file!"
                shutil.copyfile(f"{tmpdir}/gfnff_topo", self.topology)

            # restore working directory
            os.chdir(old_working_directory)

        # apply constraints
        for c in self.constraints:
            f, e = c.evaluate(positions)
            energy += e
            forces += f

        # return result
        if pipe is not None:
            assert isinstance(pipe, mp.connection.Connection), "not a valid Connection instance!"
            pipe.send([energy, forces])
            pipe.close()
        else:
            return energy, forces

class GaussianCalculator(Calculator):
    def __init__(
            self,
            charge=0,
            multiplicity=1,
            link0={"mem":"1GB", "nprocshared":"4"},
            route_card="#p hf/3-21g force",
            footer=None,
            constraints=list(),
        ):
        if not re.search("force", route_card):
            raise ValueError("need a force job to calculate forces")

        self.charge = charge
        self.multiplicity = multiplicity
        self.link0 = link0
        self.route_card = route_card
        self.footer = footer

        for c in constraints:
            assert isinstance(c, presto.constraints.Constraint), "{c} is not a valid constraint!"
        self.constraints = constraints

    def evaluate(self, atomic_numbers, positions, high_atoms=None, pipe=None, qc=False):
        """
        Gets the electronic energy and Cartesian forces for the specified geometry.

        Args:
            atomic_numbers (cctk.OneIndexedArray): the atomic numbers (int)
            positions (cctk.OneIndexedArray): the atomic positions in angstroms
            high_atoms (np.ndarray): do nothing with this
            pipe (): for multiprocessing, the connection through which objects should be returned to the parent process
            qc (bool): try quadratic convergence for tricky cases, only after default DIIS fails

        Returns:
            energy (float): in Hartree
            forces (cctk.OneIndexedArray): in amu Å per fs**2
        """
        assert shutil.which(presto.config.G16_EXEC), f"bad Gaussian executable {presto.config.G16_EXEC}"

        old_working_directory = os.getcwd()

        molecule = cctk.Molecule(atomic_numbers, positions, charge=self.charge, multiplicity=self.multiplicity)
        route_card = self.route_card
        if qc:
            logger.warning("warning: initial DIIS convergence failed, trying quadradic convergence")
            route_card = f"{self.route_card} scf=qc"

        energy, forces = None, None
        elapsed = 0
        with tempfile.TemporaryDirectory() as tmpdir:
            cctk.GaussianFile.write_molecule_to_file(f"{tmpdir}/g16-in.gjf", molecule, route_card, self.link0, self.footer)
            command = f"{presto.config.G16_EXEC} g16-in.gjf g16-out.out"

            # run g16
            try:
                start = time.time()
                result = sp.run(command, cwd=tmpdir, shell=True, capture_output=True)
                end = time.time()
                elapsed = end - start

                result.check_returncode()
                assert os.path.isfile(f"{tmpdir}/g16-out.out"), "no energy file!"
            except Exception as e:
                try:
                    assert qc is False # retry with quadradic convergence
                    self.evaluate(atomic_numbers, positions, high_atoms, pipe, qc=True)
                except Exception as e:
                    shutil.copyfile(f"{tmpdir}/g16-in.gjf", f"{old_working_directory}/g16-failed-input.gjf")
                    shutil.copyfile(f"{tmpdir}/g16-out.out", f"{old_working_directory}/g16-failed-output.out")
                    raise ValueError(f"g16 failed:\n{e}\nfiles:{os.listdir(tmpdir)}")

            # extract output
            gaussian_file = cctk.GaussianFile.read_file(f"{tmpdir}/g16-out.out")
            ensemble = gaussian_file.ensemble
            molecule = ensemble.molecules[-1]

            properties_dict = ensemble.get_properties_dict(molecule)
            energy = properties_dict["energy"]
            forces = properties_dict["forces"]
            forces = forces * presto.constants.AMU_A2_FS2_PER_HARTREE_BOHR

        # restore working directory
        os.chdir(old_working_directory)

        # apply constraints
        for c in self.constraints:
            f, e = c.evaluate(positions)
            energy += e
            forces += f

        # return result
        if pipe is not None:
            assert isinstance(pipe, mp.connection.Connection), "not a valid Connection instance!"
            pipe.send([energy, forces])
            pipe.close()
        else:
            return energy, forces

class ONIOMCalculator(Calculator):
    """
    Composes two other calculators, and computes forces according to the ONIOM embedding scheme.
    """
    def __init__(self, high_calculator, low_calculator, constraints=list()):
        assert isinstance(high_calculator, Calculator), "high calculator isn't a proper Calculator!"
        assert isinstance(low_calculator, Calculator), "low calculator isn't a proper Calculator!"
        self.high_calculator = high_calculator
        self.low_calculator = low_calculator

        for c in constraints:
            assert isinstance(c, presto.constraints.Constraint), f"{c} is not a valid constraint!"
        self.high_calculator.constraints = constraints
        self.constraints = constraints

        #### prevent namespace collisions
        self.full_calculator = copy.deepcopy(self.low_calculator)
        if isinstance(self.full_calculator, XTBCalculator):
            if self.full_calculator.gfn == "ff":
                self.full_calculator.topology = self.full_calculator.topology.replace(".low", ".full")

    def evaluate(self, atomic_numbers, positions, high_atoms, pipe=None):
        """
        Evaluates the forces according to the ONIOM embedding scheme.
        """
        assert len(high_atoms) > 0, "no point in doing ONIOM without a high layer!"
        assert isinstance(atomic_numbers, cctk.OneIndexedArray), "need to pass one-indexed array for indexing to work properly"
        assert isinstance(positions, cctk.OneIndexedArray), "need to pass one-indexed array for indexing to work properly"

        high_atomic_numbers = atomic_numbers[high_atoms]
        high_positions = positions[high_atoms]

        parent_hh, child_hh = mp.Pipe()
        process_hh = mp.Process(target=self.high_calculator.evaluate, kwargs={
            "atomic_numbers": high_atomic_numbers,
            "positions": high_positions,
            "pipe": child_hh,
        })
        process_hh.start()

        parent_hl, child_hl = mp.Pipe()
        process_hl = mp.Process(target=self.low_calculator.evaluate, kwargs={
            "atomic_numbers": high_atomic_numbers,
            "positions": high_positions,
            "pipe": child_hl,
        })
        process_hl.start()

        parent_ll, child_ll = mp.Pipe()
        process_ll = mp.Process(target=self.full_calculator.evaluate, kwargs={
            "atomic_numbers": atomic_numbers,
            "positions": positions,
            "pipe": child_ll,
        })
        process_ll.start()

        e_hh, f_hh = parent_hh.recv()
        e_hl, f_hl = parent_hl.recv()
        e_ll, f_ll = parent_ll.recv()

        process_hh.join()
        process_hl.join()
        process_ll.join()

        energy = e_hh + e_ll - e_hl
        forces = f_ll
        forces[high_atoms] = f_hh + f_ll[high_atoms] - f_hl

        return energy, forces

def build_calculator(settings, checkpoint_filename, constraints=list(), ):
    """
    Build calculator from settings dict.
    """
    assert isinstance(settings, dict), "Need to pass a dictionary!!"
    assert "type" in settings, "Need `type` for calculator"
    assert isinstance(settings["type"], str), "Calculator `type` must be a string"

    if settings["type"].lower() == "oniom":
        assert "high_calculator" in settings, "Need `high_calculator` settings dictionary for ONIOM!"
        assert "low_calculator" in settings, "Need `low_calculator` settings dictionary for ONIOM!"
        return ONIOMCalculator(
            high_calculator=build_calculator(settings["high_calculator"], checkpoint_filename + ".high"),
            low_calculator=build_calculator(settings["low_calculator"], checkpoint_filename + ".low"),
            constraints=constraints,
        )
    elif settings["type"].lower() == "gaussian":
        charge = 0
        if "charge" in settings:
            assert isinstance(settings["charge"], int), "Calculator `charge` must be an integer."
            charge = settings["charge"]

        multiplicity = 1
        if "multiplicity" in settings:
            assert isinstance(settings["multiplicity"], int), "Calculator `multiplicity` must be an integer."
            assert settings["multiplicity"] > 0, "Calculator `multiplicity` must be positive."
            multiplicity = settings["multiplicity"]

        link0 = None
        if "link0" in settings:
            assert isinstance(settings["link0"], dict), "Calculator `link0` must be a dictionary."
            link0 = settings["link0"]

        assert "route_card" in settings, "Need a route card for `GaussianCalculator`!"
        assert isinstance(settings["route_card"], str), "Calculator `route_card` must be a string."
        assert re.search("#p", settings["route_card"]), "Need `#p` in route card!"
        assert re.search("force", settings["route_card"]), "Need `force` in route card!"

        footer = None
        if "footer" in settings:
            assert isinstance(settings["footer"], str), "Calculator `footer` must be string or `None`."
            footer = settings["footer"]

        if link0 is not None:
            return GaussianCalculator(charge=charge, multiplicity=multiplicity, link0=link0, route_card=settings["route_card"], footer=footer, constraints=constraints)
        else:
            return GaussianCalculator(charge=charge, multiplicity=multiplicity, route_card=settings["route_card"], footer=footer, constraints=constraints)

    elif settings["type"].lower() == "xtb":
        charge = 0
        if "charge" in settings:
            assert isinstance(settings["charge"], int), "Calculator `charge` must be an integer."
            charge = settings["charge"]

        multiplicity = 1
        if "multiplicity" in settings:
            assert isinstance(settings["multiplicity"], int), "Calculator `multiplicity` must be an integer."
            assert settings["multiplicity"] > 0, "Calculator `multiplicity` must be positive."
            multiplicity = settings["multiplicity"]

        gfn = 2
        if "gfn" in settings:
            assert settings["gfn"] in [0, 1, 2, "ff"], "Calculator `gfn` must be 0, 1, 2, or ``ff``."
            gfn = settings["gfn"]

        parallel = 1
        if "parallel" in settings:
            assert isinstance(settings["parallel"], int), "Calculator `parallel` must be an integer."
            assert settings["parallel"] > 0, "Calculator `parallel` must be positive."
            parallel = settings["parallel"]

        xcontrol = None
        if "xcontrol" in settings:
            assert isinstance(settings["xcontrol"], str), "Calculator `xcontrol` must be a string."
            xcontrol = settings["xcontrol"]

        topology = None
        if "topology" in settings:
            assert isinstance(settings["topology"], str), "Calculator `topology` must be a string (path where topology will be stored)."
            topology = settings["topology"]
        elif gfn == "ff":
            topology = f"{checkpoint_filename}.top"

        return XTBCalculator(charge=charge, multiplicity=multiplicity, gfn=gfn, parallel=parallel, constraints=constraints, xcontrol_path=xcontrol, topology=topology)

    else:
        raise ValueError(f"Unknown integrator type {settings['type']}! Allowed options are `oniom`, `xtb`, or `gaussian`.")





import presto, logging, cctk
import numpy as np

logger = logging.getLogger(__name__)

class Check():
    """
    Class for recurring trajectory check.
    """
    def __init__(self, interval=20):
        assert isinstance(interval, int), "interval must be integer"
        self.interval = interval

    def check(self, trajectory):
        """
        Checks a trajectory. Will raise an exception if anything is wrong.

        Args:
            trajectory (presto.trajectory.Trajectory): trajectory to check

        Returns:
            nothing
        """
        pass

class VelocityCheck(Check):
    """
    Checks that velocities stay sane.

    Attributes:
        interval (int): interval to check at
        max_velocity (float): velocity not to exceed (per component).
    """
    def __init__(self, interval=20, max_velocity=1, **kwargs):
        assert isinstance(interval, int), "interval must be integer"
        self.interval = interval

        assert isinstance(max_velocity, (int, float)), "max_velocity must be numeric"
        self.max_velocity = max_velocity

    def check(self, trajectory):
        frame = trajectory.frames[-1]
        assert np.max(frame.velocities) < self.max_velocity, f"VelocityCheck failed! max allowed: {self.max_velocity:.2f}, max obtained; {np.max(frame.velocities):.2f}"

class RadiusCheck(Check):
    """
    Checks that particles do not go past a set radius. By default, there is a 10% buffer past the applied potential.

    Attributes:
        interval (int): interval to check at
        buffer (float): multiplied by ``potential.max_radius`` to give max allowable radius
    """
    def __init__(self, interval=20, buffer=1.1, **kwargs):
        assert isinstance(interval, int), "interval must be integer"
        self.interval = interval

        assert isinstance(buffer, (int, float)), "buffer must be numeric"
        self.buffer = buffer

    def check(self, trajectory):
        frame = trajectory.frames[-1]
        positions = frame.positions.view(np.ndarray)
        radii = np.linalg.norm(frame.positions, axis=1)
        max_r = trajectory.integrator.potential.max_radius * self.buffer
        assert np.max(radii) < max_r, f"RadiusCheck failed! max allowed: {max_r:.2f}, max obtained; {np.max(radii):.2f}"

class TopologyCheck(Check):
    """
    Checks that the frame has the same topology as it did ``interval`` frames ago.

    Attributes:
        interval (int): interval to check at
    """
    def __init__(self, interval=100, **kwargs):
        super().__init__(interval)

    def check(self, trajectory):
        new_frame = trajectory.frames[-1]
        old_frame = trajectory.frames[-1*(self.interval+1)]
        new_mol = new_frame.molecule().assign_connectivity()
        old_mol = old_frame.molecule().assign_connectivity()
        assert cctk.topology.are_isomorphic(new_mol, old_mol), "TopologyCheck failed! different topologies found."

def build_check(settings):
    if settings["type"] == "velocity":
        return VelocityCheck(**settings)
    elif settings["type"] == "radius":
        return RadiusCheck(**settings)
    elif settings["type"] == "topology":
        return TopologyCheck(**settings)
    else:
        raise ValueError(f"unknown presto.check.Check type {settings['type']}!")



import unittest, cctk
import numpy as np

import sys
sys.path.append('../presto')

import presto
from presto import calculators

if __name__ == '__main__':
    unittest.main()

class TestONIOM(unittest.TestCase):
    def test_simple(self):
        hydrogen_molecule = cctk.XYZFile.read_file("test/static/H2.xyz").molecule
        atomic_numbers = hydrogen_molecule.atomic_numbers
        positions = hydrogen_molecule.geometry
        high_atoms = np.array([1])
        xtb_calculator = calculators.XTBCalculator()
        oniom_calculator = calculators.ONIOMCalculator(high_calculator=xtb_calculator, low_calculator=xtb_calculator)

        energy, forces = oniom_calculator.evaluate(atomic_numbers, positions, high_atoms)
        self.assertLessEqual(abs(energy+0.90379671599), 0.00000001)
        expected_forces = [[ 0.21688017, 0, 0],[-0.21688017, 0, 0]]
        actual_forces = [ list(i) for i in forces ]
        for row1,row2 in zip(expected_forces, actual_forces):
            for entry1,entry2 in zip(row1, row2):
                delta = abs(entry1-entry2)
                self.assertLessEqual(delta, 0.00000001)



import unittest, cctk
import numpy as np

import sys
sys.path.append('../presto')

import presto

if __name__ == '__main__':
    unittest.main()

class TestFrame(unittest.TestCase):
    def gen_test_frame(self):
        zs = cctk.OneIndexedArray([2])
        traj = presto.trajectory.Trajectory(
            timestep=0.5,
            atomic_numbers=zs,
            high_atoms=np.array([]),
            active_atoms=np.array([1]),
            calculator=presto.calculators.XTBCalculator(),
            integrator=presto.integrators.VelocityVerletIntegrator(),
            stop_time=10000,
        )

        self.assertTrue(isinstance(traj, presto.trajectory.Trajectory))

        e0 = cctk.OneIndexedArray([[0,0,0]])
        e1 = cctk.OneIndexedArray([[1,0,0]])
        frame = presto.frame.Frame(traj, e0, e1, e0)
        self.assertTrue(isinstance(frame, presto.frame.Frame))
        return frame

    def test_temp(self):
        frame = self.gen_test_frame()
        self.assertEqual(int(frame.temperature()), 1604675)

    def test_next(self):
        frame = self.gen_test_frame()
        f2 = frame.next()
        self.assertTrue(isinstance(f2, presto.frame.Frame))
        self.assertListEqual(list(f2.positions[1]), [0.5, 0, 0])

        f3 = f2.next()
        self.assertListEqual(list(f3.positions[1]), [1, 0, 0])

        f4 = f3.next()
        self.assertListEqual(list(f4.positions[1]), [1.5, 0, 0])

        f5 = f4.next()
        self.assertListEqual(list(f5.positions[1]), [2, 0, 0])

        f0 = frame.prev()
        self.assertListEqual(list(f0.positions[1]), [-0.5, 0, 0])

    def test_com_motion(self):
        num_atoms = 4
        zs = cctk.OneIndexedArray([1] * num_atoms)
        traj = presto.trajectory.Trajectory(
            timestep=0.5,
            atomic_numbers=zs,
            high_atoms=np.array([]),
            active_atoms=np.array([]),
            calculator=presto.calculators.XTBCalculator(),
            integrator=presto.integrators.VelocityVerletIntegrator(),
            stop_time = 1000,
        )

        positions = cctk.OneIndexedArray([[2,0,0],[-1,0,0],[0,1,0],[0,-1,0]])
        velocities = cctk.OneIndexedArray([[0,2,1],[0,-1,0],[1,0,-2],[-1,0,0]])
        accels = cctk.OneIndexedArray([[0,0,0],[0,0,0],[0,0,0],[0,0,0]])

        frame = presto.frame.Frame(traj, positions, velocities, accels)
        frame.remove_com_motion()
        self.assertTrue(np.linalg.norm(np.sum(traj.masses.reshape(-1,1) * frame.velocities, axis=0)) < 0.0001)



import numpy as np
import cctk, math
import matplotlib.pyplot as plt

import presto

def autocorrelation(values):

    block_length = 1
    variance = 0

    return block_length, variance

def radial_distribution(frames, indices, cutoff=10, center_cutoff=5, resolution=0.025):
    """
    Args:
        frames (list of presto.Frame):
        indices (list): list of indices to analyze from each frame
        cutoff (float): max distance to examine
        center_cutoff (float): max distance from center for starting atoms)
        resolution (float): bin size

    Returns:
        list of counts at each point divided by volume
        list of shell radii
    """

    assert (100 * cutoff) % (100 * resolution) == 0, "cutoff must be a multiple of resolution!"
    #### floating point accuracy issues

    count = 0
    frame_count = 0
    total_hist = np.zeros(shape=int(cutoff/resolution))

    radii = np.arange(0, cutoff, resolution)

    for frame in frames:
        if frame_count % 100 == 0:
            print(f"new frame! (#{frame_count}) {count} counts")
        frame_count += 1

        positions = frame.positions[indices].view(np.ndarray)
        for x1 in positions:
            if np.linalg.norm(x1 > 5):
                continue
            for x2 in positions:
                if np.array_equal(x1, x2):
                    continue
                distance = np.linalg.norm(x1 - x2)
                print(f"observed distance {distance}")

                count += 1
                index = int(distance/resolution)
                if index < len(total_hist):
                    print(f"assigning index {index} and radius {radii[index]}")
                    total_hist[index] += 1

    volumes_per_shell = 4/3 * math.pi * (np.power(radii + resolution, 3) - np.power(radii, 3))

    vol_correction = frame.volume() / volumes_per_shell

    g = total_hist * vol_correction / (frame_count * 0.5 * len(indices) * (len(indices) - 1))
    return g, radii



import unittest, cctk
import numpy as np

import sys, shutil
sys.path.append('../presto')

import presto
from presto import calculators

if __name__ == '__main__':
    unittest.main()

class TestGaussian(unittest.TestCase):
    def test_gaussian(self):
        if not shutil.which("g16"):
            return True

        hydrogen_molecule = cctk.XYZFile.read_file("test/static/H2.xyz").molecule
        atomic_numbers = hydrogen_molecule.atomic_numbers
        positions = hydrogen_molecule.geometry
        gaussian_calculator = calculators.GaussianCalculator(
            charge=0,
            multiplicity=1,
            link0={"mem":"1GB", "nprocshared":"1"},
            route_card = "#p hf 3-21g force",
            footer = None
        )
        energy, forces = gaussian_calculator.evaluate(atomic_numbers, positions)
        self.assertLessEqual(abs(energy+1.0508014119), 0.00000001)
        expected_forces = [[ 0.21550274, 0, 0],[-0.21550274, 0, 0]]
        actual_forces = [ list(i) for i in forces ]
        for row1,row2 in zip(expected_forces, actual_forces):
            for entry1,entry2 in zip(row1, row2):
                delta = abs(entry1-entry2)
                self.assertLessEqual(delta, 0.00000001)



import unittest, cctk
import numpy as np
from datetime import datetime

import sys, os
sys.path.append('../presto')

import presto
from asciichartpy import plot

if __name__ == '__main__':
    unittest.main()

class TestReplicaExchange(unittest.TestCase):
    def test_remd(self):
        if os.path.exists("test/static/remd-chk1.chk"):
            os.remove("test/static/remd-chk1.chk")
        if os.path.exists("test/static/remd-chk2.chk"):
            os.remove("test/static/remd-chk2.chk")
        def sched_300k(time):
            return 300

        def sched_330k(time):
            return 330

        zs = cctk.OneIndexedArray([2, 10])
        traj1 = presto.trajectory.EquilibrationTrajectory(
            timestep=1,
            atomic_numbers=zs,
            high_atoms=np.array([2]),
            active_atoms=np.array([1,2]),
            calculator=presto.calculators.XTBCalculator(),
            integrator=presto.integrators.LangevinIntegrator(),
            bath_scheduler=sched_300k,
            checkpoint_filename="test/static/remd-chk1.chk",
            stop_time = 400,
        )

        traj2 = presto.trajectory.EquilibrationTrajectory(
            timestep=1,
            atomic_numbers=zs,
            high_atoms=np.array([2]),
            active_atoms=np.array([1,2]),
            calculator=presto.calculators.XTBCalculator(),
            integrator=presto.integrators.LangevinIntegrator(),
            bath_scheduler=sched_330k,
            checkpoint_filename="test/static/remd-chk2.chk",
            stop_time = 400,
        )

        x = cctk.OneIndexedArray([[1,0,0], [0,0,0]])
        traj1.initialize(positions=x)
        x = cctk.OneIndexedArray([[2,0,0], [0,0,0]])
        traj2.initialize(positions=x)

        remd = presto.replica_exchange.ReplicaExchange(trajectories=[traj1, traj2], swap_interval=25, checkpoint_filename="test/static/remd.chk")

        remd.run()




import presto, logging
import numpy as np

logger = logging.getLogger(__name__)

class Reporter():
    """
    Class for recurring report action.
    """
    def __init__(self, interval=10):
        assert isinstance(interval, int), "interval must be integer"
        self.interval = interval

    def report(self, trajectory):
        """
        Writes report (usually to external file.)

        Args:
            frame (presto.frame.Frame): frame to check

        Returns:
            nothing
        """
        pass

class TimingReporter(Reporter):
    def report(self, trajectory):
        times = [f.elapsed for f in trajectory.frames[-self.interval:]]
        print("AVERAGE TIME PER FRAME:")
        print(np.average(times))

class PropertyReporter(Reporter):
    def __init__(self, interval=10, properties=list()):
        assert isinstance(interval, int), "interval must be integer"
        self.interval = interval

    def report(self, trajectory):
        pass



import unittest, cctk
import numpy as np

import sys, os
sys.path.append('../presto')

import presto
from presto import calculators

if __name__ == '__main__':
    unittest.main()

class TestXTB(unittest.TestCase):
    def test_gfn2(self):
        hydrogen_molecule = cctk.XYZFile.read_file("test/static/H2.xyz").molecule
        atomic_numbers = hydrogen_molecule.atomic_numbers
        positions = hydrogen_molecule.geometry
        xtb_calculator = calculators.XTBCalculator(charge=0, multiplicity=1)
        energy, forces = xtb_calculator.evaluate(atomic_numbers, positions, )
        self.assertLessEqual(abs(energy+0.90379671599), 0.00000001)
        expected_forces = [[ 0.21688017, 0, 0],[-0.21688017, 0, 0]]
        actual_forces = [ list(i) for i in forces ]
        for row1,row2 in zip(expected_forces, actual_forces):
            for entry1,entry2 in zip(row1, row2):
                delta = abs(entry1-entry2)
                self.assertLessEqual(delta, 0.00000001)

    def test_gfn0(self):
        hydrogen_molecule = cctk.XYZFile.read_file("test/static/H2.xyz").molecule
        atomic_numbers = hydrogen_molecule.atomic_numbers
        positions = hydrogen_molecule.geometry
        xtb_calculator = calculators.XTBCalculator(charge=0, multiplicity=1, gfn=0)
        energy, forces = xtb_calculator.evaluate(atomic_numbers, positions, )
        self.assertLessEqual(abs(energy+1.0083914049), 0.00000001)

    def test_big(self):
        molecule = cctk.XYZFile.read_file("test/static/nazarov-elim-solvated.xyz").molecule
        atomic_numbers = molecule.atomic_numbers
        positions = molecule.geometry
        xtb_calculator = calculators.XTBCalculator(charge=0, multiplicity=1, gfn="ff", parallel=4, topology="test/static/nazarov-elim.top")
        energy, forces = xtb_calculator.evaluate(atomic_numbers, positions, )
        self.assertLessEqual(abs(energy+122.8153475072), 0.00000001)



import numpy as np
import cctk
import presto

class Potential():
    """
    Constraining potential to keep things confined.

    Attributes:
        max_radius (float): maximum allowed distance for particles. for a sphere, this is just the radius. for a cube, it might be half the space diagonal.
            for an ellipsoid, it might be the major radius. used for internal automatic checks.
    """
    def evaluate(self, positions):
        """
        Returns energy, forces from a given set of atomic coordinates.

        Args:
            positions (cctk.OneIndexedArray):

        Returns:
            energy (float):
            forces (cctk.OneIndexedArray):
        """
        pass

class SphericalHarmonicPotential(Potential):
    """
    Default force constant is 10 kcal/mol per Å**2, akin to that employed by Singleton (JACS, 2016, 138, 15167).
    In *presto* units, this is 0.004184 amu Å**2 fs**-2.

    Attributes:
        radius (float): area outside which this takes effect
        force_constant (float):
        convert_from_kcal (bool):
    """
    def __init__(self, radius, force_constant=10, convert_from_kcal=True):
        assert isinstance(radius, (int, float)), "radius must be numeric"
        assert radius > 0, "radius must be positive"

        assert isinstance(force_constant, (int, float)), "force_constant must be numeric"
        assert force_constant > 0, "force_constant must be positive"
        if convert_from_kcal:
            force_constant *= 0.0004184

        self.radius = radius
        self.force_constant = force_constant

        self.max_radius = radius

    def __str__(self):
        return f"SphericalHarmonicPotential(radius={self.radius:.2f},force_constant={self.force_constant:.5f})"

    def __repr__(self):
        return f"SphericalHarmonicPotential(radius={self.radius:.2f},force_constant={self.force_constant:.5f})"

    def evaluate(self, positions):
        radii = np.linalg.norm(positions, axis=1)
        # distance from equilibrium
        excess_radius = positions - positions/radii.reshape(-1,1) * self.radius

        # F = -k * x
        forces = -1 * self.force_constant * excess_radius

        # E = 0.5 * k * x**2
        energies = 0.5 * self.force_constant * np.linalg.norm(excess_radius, axis=1) ** 2

        # it's a one-sided spring.
        inside = radii < self.radius
        forces.view(np.ndarray)[inside,:] = 0
        energies.view(np.ndarray)[inside] = 0

        return np.sum(energies), forces.view(cctk.OneIndexedArray)

def build_potential(settings):
    """
    Build potential from settings dict.
    """
    assert isinstance(settings, dict), "Need to pass a dictionary!!"
    assert "type" in settings, "Need `type` for potential"
    assert isinstance(settings["type"], str), "Potential `type` must be a string"

    if settings["type"].lower() == "spherical_harmonic":
        assert "radius" in settings, "Need `radius` for spherical harmonic potential."
        assert isinstance(settings["radius"], (int, float)), "`radius` must be numeric!"
        assert settings["radius"] > 0, "`radius` must be positive!"

        if "force_constant" in settings:
            assert isinstance(settings["force_constant"], (int, float)), "`force_constant` must be numeric!"
            assert settings["force_constant"] > 0, "`force_constant` must be positive!"
            return SphericalHarmonicPotential(radius=settings["radius"], force_constant=settings["force_constant"])
        else:
            return SphericalHarmonicPotential(radius=settings["radius"])

    else:
        raise ValueError(f"Unknown potential type {settings['type']}! Allowed options are `spherical_harmonic` (free will is an illusion).")






import unittest, cctk, os
import numpy as np

import sys
sys.path.append('../presto')
import presto

if __name__ == '__main__':
    unittest.main()

class TestConfig(unittest.TestCase):
    def test_load_config(self):
        path = "test/static/ex.yaml"
        presto.config.build(path, "test/static/ex.chk", "test/static/nazarov-elim-solvated.xyz")

        path = "test/static/ex2.yaml"
        presto.config.build(path, "test/static/ex2.chk", oldchk="test/static/ex.chk")

        path = "test/static/ex3.yaml"
        presto.config.build(path, "test/static/ex3.chk")

        os.remove("test/static/ex.chk")
        os.remove("test/static/ex2.chk")
        os.remove("test/static/ex3.chk")




import numpy as np
import math, copy, cctk, os, re, logging, time, yaml, random
import multiprocessing as mp

import presto

logger = logging.getLogger(__name__)

MIN_CHECKPOINT_INTERVAL = 50

class ReplicaExchange():
    """
    Runs several trajectories and manages interconversions between them.

    Based on:
        http://www.math.pitt.edu/~cbsg/Materials/Earl_ParallelTempering.pdf

    Attributes:
        stop_time (int):
        swaps (list of dict):
        swap_interval (int):
        checkpoint_filename (int):
        finished (bool):
        current_idx (int):
    """
    def __init__(self, trajectories, checkpoint_filename="remd.chk", swap_interval=10):
        temps = np.zeros(shape=len(trajectories))
        for idx, traj in enumerate(trajectories):
            assert isinstance(traj, presto.trajectory.EquilibrationTrajectory), "all trajectories must be EquilibrationTrajectories"
            assert traj.timestep == trajectories[0].timestep, "all trajectories must have same ``timestep``"
            assert traj.stop_time == trajectories[0].stop_time, "all trajectories must have same ``stop_time``"
            assert np.array_equal(traj.high_atoms, trajectories[0].high_atoms), "all trajectories must have same ``high_atoms``"
            assert np.array_equal(traj.active_atoms, trajectories[0].active_atoms), "all trajectories must have same ``active_atoms``"

        self.trajectories = sorted(trajectories, key=lambda x: x.bath_scheduler(0))
        self.stop_time = trajectories[0].stop_time

        self.swaps = list()
        assert isinstance(swap_interval, (float, int))
        assert swap_interval % trajectories[0].timestep == 0, "swap_interval must be a multiple of timestep"
        self.swap_interval = swap_interval

        assert isinstance(checkpoint_filename, str)
        self.checkpoint_filename = checkpoint_filename
        self.finished = False
        self.current_idx = 0

        self.load()

    def __str__(self):
        return f"ReplicaExchange({len(trajectories)} trajectories, swap_interval={swap_interval} fs, checkpoint_file={checkpoint_filename})"

    def __repr__(self):
        return f"ReplicaExchange({len(trajectories)} trajectories, swap_interval={swap_interval} fs, checkpoint_file={checkpoint_filename})"

    def save(self):
        file_dict = {
            "stop_time": self.stop_time,
            "swap_interval": self.swap_interval,
            "swaps": self.swaps,
            "finished": self.finished,
            "current_idx": self.current_idx,
        }
        with open(self.checkpoint_filename, "w+") as f:
            yaml.dump(file_dict, f)
        logger.info(f"Saving ReplicaExchange to {self.checkpoint_filename}.")
        return self

    def load(self):
        if os.path.exists(self.checkpoint_filename):
            with open(self.checkpoint_filename, "r+") as f:
                file_dict = yaml.safe_load(f)
                self.stop_time = file_dict["stop_time"]
                self.swap_interval = file_dict["swap_interval"]
                self.swaps = file_dict["swaps"]
                self.finished = file_dict["finished"]
                self.current_idx = file_dict["current_idx"]
            logger.info(f"Loaded ReplicaExchange from {self.checkpoint_filename}.")
        return self

    def run(self):
        if self.finished:
            logger.info(f"Replica exchange already finished!")
            return self

        processes = [None]*len(self.trajectories)
        start_idx = self.current_idx

        # we break the runs up into small chunks
        for current_idx in range(start_idx, int(self.stop_time/self.swap_interval)):
            next_idx = current_idx + 1
            target_time = next_idx * self.swap_interval
            for idx, traj in enumerate(self.trajectories):
                # how long does each traj need to run for?
                time_remaining = max(0, target_time - traj.last_time_run())

                if time_remaining:
                    processes[idx] = mp.Process(target=traj.run, kwargs={
                        "time": self.swap_interval,
                        "checkpoint_interval": min(MIN_CHECKPOINT_INTERVAL, self.swap_interval),
                    })
                    processes[idx].start()

            for process in processes:
                if process is not None:
                    process.join()

            for traj in self.trajectories:
                traj.load_from_checkpoint()
                assert traj.last_time_run() == target_time

            self.exchange(next_idx*self.swap_interval)
            self.current_idx = next_idx
            self.save()
        self.finished = True
        return self

    def exchange(self, time):
        kB = presto.constants.BOLTZMANN_CONSTANT

        for i in range(len(self.trajectories)-1):
            j = i+1

            E_i = self.trajectories[i].frames[-2].energy
            E_j = self.trajectories[j].frames[-2].energy
            T_i = self.trajectories[i].bath_scheduler(0)
            T_j = self.trajectories[j].bath_scheduler(0)
            b_i = 1 / (kB*T_i)
            b_j = 1 / (kB*T_j)

            p = min(1, np.exp( (E_i - E_j) * (b_i - b_j) ) )
            logger.info(f"E{i} & E{j}\tp={p}")

            if p > random.random():
                v_i_scaling = np.sqrt(T_j/T_i)
                v_j_scaling = np.sqrt(T_i/T_j)

                frame_i = self.trajectories[i].frames[-1]
                frame_j = self.trajectories[j].frames[-1]
                frame_i.velocities *= v_i_scaling
                frame_j.velocities *= v_j_scaling

                self.trajectories[i].frames[-1] = frame_j
                self.trajectories[j].frames[-1] = frame_i

                self.swaps.append({"time": time, "i": i, "j": j})
                logger.info(f"\nReplicas {i} & {j} swapped after {time} fs!\t{E_i} {E_j} p={p}")

    def report(self):
        counts = np.zeros(shape=len(self.trajectories))
        possible = self.current_idx * (len(self.trajectories) - 1)

        for swap in self.swaps:
            counts[swap["i"]] += 1

        text = f"Time: {self.current_idx * self.swap_interval} fs\n"
        text += f"Exchange probability: {np.sum(counts)/possible:.2%}\n"
        for i in range(len(self.trajectories) - 1):
            text += f"\tReplica {i} <=> Replica {i+1}  \t{counts[i]/self.current_idx:.2%}\n"

        return text



import numpy as np
import math, copy, cctk, os, re, logging, time
import fasteners

import h5py
import presto

logger = logging.getLogger(__name__)

class Trajectory():
    """

    Attributes:
        timestep (float): in fs
        frames (list of presto.Frame):
        stop_time (float): how long to run for

        high_atoms (np.ndarray): to calculate at high level of theory, list of 1-indexed atom numbers
        active_atoms (np.ndarray): non-frozen atoms, list of 1-indexed atom numbers

        atomic_numbers (cctk.OneIndexedArray): list of atomic numbers
        masses (cctk.OneIndexedArray): list of masses

        calculator (presto.calculators.Calculator):
        integrator (presto.integrators.Integrator):
        reporters (list of presto.reporters.Reporter):
        checks (list of presto.checks.Check):

        finished (bool):
        forwards (bool):

        checkpoint_filename (str):
        lock (fasteners.InterProcessLock): lock object
        save_interval (int): how many frames to save
    """

    def __init__(
        self,
        calculator=None,
        integrator=None,
        reporters=list(),
        checks=list(),
        timestep=None,
        atomic_numbers=None,
        high_atoms=None,
        forwards=True,
        checkpoint_filename=None,
        stop_time=None,
        save_interval=1,
        **kwargs
    ):

        # do this first!
        if timestep is not None:
            assert timestep > 0, "can't have timestep ≤ 0!"
            self.timestep = float(timestep)

        if checkpoint_filename is not None:
            assert isinstance(checkpoint_filename, str), "need string for file"
        self.checkpoint_filename = checkpoint_filename

        self.lock = None
        self.initialize_lock()
        self.frames = list()

        if self.has_checkpoint():
            if "load_frames" in kwargs:
                self.load_from_checkpoint(kwargs["load_frames"])
            else:
                self.load_from_checkpoint(slice(-1,None,None))

        if calculator is not None:
            assert isinstance(calculator, presto.calculators.Calculator), "need a valid calculator!"
        self.calculator = calculator

        if integrator is not None:
            assert isinstance(integrator, presto.integrators.Integrator), "need a valid integrator!"
        self.integrator = integrator

        assert all([isinstance(c, presto.checks.Check) for c in checks])
        self.checks = checks

        assert all([isinstance(r, presto.reporters.Reporter) for r in reporters])
        self.reporters = reporters

        if atomic_numbers is not None:
            assert isinstance(atomic_numbers, cctk.OneIndexedArray), "atomic numbers must be cctk 1-indexed array!"
            self.atomic_numbers = atomic_numbers
        elif not hasattr(self, "atomic_numbers"):
            raise ValueError("no atomic numbers specified")

        if not hasattr(self, "finished"):
            self.finished = False

        if high_atoms is not None:
            assert isinstance(high_atoms, np.ndarray), "high_atoms must be np.ndarray!"
            self.high_atoms = high_atoms
        else:
            self.high_atoms = None

        active_atoms = None
        if "active_atoms" in kwargs:
            active_atoms = kwargs["active_atoms"]
            assert isinstance(active_atoms, np.ndarray), "active_atoms must be np.ndarray!"
            self.active_atoms = active_atoms
        elif "inactive_atoms" in kwargs:
            self.set_inactive_atoms(kwargs["inactive_atoms"])
        else:
            # assume all atoms are active
            self.set_inactive_atoms(np.ndarray([]))

        if not hasattr(self, "masses"):
            self.masses = cctk.OneIndexedArray([float(cctk.helper_functions.draw_isotopologue(z)) for z in atomic_numbers])

        if not hasattr(self, "frames"):
            self.frames = []

        if forwards is not None:
            assert isinstance(forwards, bool), "forwards must be bool"
            self.forwards = forwards
        elif not hasattr(self, "forwards"):
            self.forwards = True

        if not hasattr(self, "stop_time"):
            assert (isinstance(stop_time, float)) or (isinstance(stop_time, int)), "stop_time needs to be numeric!"
            assert stop_time > 0, "stop_time needs to be positive!"
            self.stop_time = stop_time

        assert isinstance(save_interval, int), "save_interval needs to be positive"
        assert save_interval > 0, "save_interval needs to be positive"
        self.save_interval = save_interval

    def __str__(self):
        return f"Trajectory({len(self.frames)} frames)"

    def __repr__(self):
        return f"Trajectory({len(self.frames)} frames)"

    def set_inactive_atoms(self, inactive_atoms):
        """
        Since sometimes it's easier to specify the inactive atoms than the inactive atoms, this method updates ``self.active_atoms`` with the complement of ``inactive_atoms``.
        """
        assert isinstance(inactive_atoms, (list, np.ndarray)), "Need list of atoms!"
        active_atoms = list(range(1, len(self.atomic_numbers)+1))
        if len(inactive_atoms):
            for atom in inactive_atoms:
                active_atoms.remove(atom)
        active_atoms = np.array(active_atoms)
        self.active_atoms = active_atoms

    def run(self, checkpoint_interval=10, keep_all=False, time=None, **kwargs):
        """
        Run the trajectory.

        Args:
            checkpoint_interval (int): interval at which to save (in frames, not fs)
            keep_all (bool): whether or not to keep all frames in memory
            time (float): total time to run for -- default is None, implying trajectory should be run until finished
        """
        if self.checkpoint_filename is None:
            if "checkpoint_filename" in kwargs:
                self.checkpoint_filename = kwargs["checkpoint_filename"]
            else:
                raise ValueError("no checkpoint filename given!")

        self.load_from_checkpoint(slice(-1, None, None))
        assert len(self.frames) == 1, "Wrong number of frames - do you need to call trajectory.initialize()?"

        if self.finished:
            logger.info("Trajectory already finished!")
        else:
            # initialize runtime controller
            controller = presto.controller.Controller(self, **kwargs)
            try:
                controller.run(checkpoint_interval=checkpoint_interval, runtime=time)
            except Exception as e:
                raise ValueError(f"Trajectory run terminated prematurely due to error: {e}")

        if keep_all:
            self.load_from_checkpoint()
            assert self.frames[0].time == 0, "missing first frame despite keep_all being True!"

        return self

    def initialize(self):
        """
        Adds first frame with randomly-initialized velocities. Should call ``self.save()``.
        """
        pass

    def has_checkpoint(self):
        if self.checkpoint_filename is None:
            return False
        if os.path.exists(self.checkpoint_filename):
            return True
        else:
            return False

    def load_from_checkpoint(self, frames=slice(None)):
        """
        Loads frames from ``self.checkpoint_filename``.

        Args:
            frames (Slice object): if not all frames are desired, a Slice object can be passed

        Returns:
            nothing
        """
        if not self.has_checkpoint():
            return # nothing to load!

        self.initialize_lock()
        self.lock.acquire()

        with h5py.File(self.checkpoint_filename, "r") as h5:
            atomic_numbers = h5.attrs["atomic_numbers"]
            self.atomic_numbers = cctk.OneIndexedArray(atomic_numbers)

            masses = h5.attrs["masses"]
            self.masses = cctk.OneIndexedArray(masses)

            self.finished = h5.attrs['finished']
            self.forwards = h5.attrs['forwards']

            self.frames = []
            if len(h5.get("all_energies")):
                all_energies = h5.get("all_energies")[frames]
                all_positions = h5.get("all_positions")[frames]
                all_velocities= h5.get("all_velocities")[frames]
                all_accels = h5.get("all_accelerations")[frames]
                temperatures = h5.get("bath_temperatures")[frames]

                # v0.2.2 - provisionally removing this
#                all_times = None
#                try:
                all_times = h5.get("all_times")[frames]
#                except Exception as e:
#                    all_times = np.arange(0, self.timestep*len(all_energies)*self.save_interval, self.timestep*self.save_interval)
#                    # this was added recently, so may be some backwards compatibility issues.
#                    pass

                if isinstance(all_energies, np.ndarray):
                    assert len(all_positions) == len(all_energies)
                    assert len(all_velocities) == len(all_energies)
                    assert len(all_accels) == len(all_energies)
                    assert len(all_times) == len(all_energies)

                for i, t in enumerate(all_times):
                    self.frames.append(presto.frame.Frame(
                        self,
                        all_positions[i].view(cctk.OneIndexedArray),
                        all_velocities[i].view(cctk.OneIndexedArray),
                        all_accels[i].view(cctk.OneIndexedArray),
                        energy=all_energies[i],
                        bath_temperature=temperatures[i],
                        time=all_times[i],
                    ))

        logger.info(f"Loaded trajectory from checkpoint file {self.checkpoint_filename} -- {len(self.frames)} frames read.")

        self.lock.release()
        return

    def num_frames(self):
        if self.has_checkpoint():
            num = 0
            with h5py.File(self.checkpoint_filename, "r") as h5:
                num = len(h5.get("all_energies"))
            return num
        else:
            return len(self.frames)

    def save(self, keep_all=False):
        if self.checkpoint_filename is None:
            raise ValueError("can't save without checkpoint filename")
        self.initialize_lock()
        self.lock.acquire()

        last_run_time = self.frames[-1].time
        if self.has_checkpoint():
            with h5py.File(self.checkpoint_filename, "r+") as h5:
                n_atoms = len(self.atomic_numbers)
                h5.attrs['finished'] = self.finished

                all_energies = h5.get("all_energies")
                old_n_frames = len(all_energies)
                # v0.2.2 - provisionally removing this
#                if "all_times" not in h5:
#                    # time is a new column, so old checkpoints may not have it.
#                    old_times = np.arange(0, self.timestep, self.timestep*old_n_frames)
#                    h5.create_dataset("all_times", data=old_times, maxshape=(None,), compression="gzip", compression_opts=9)

                all_times = h5.get("all_times")
                last_saved_time = all_times[-1]
                new_n_frames = int((last_run_time - last_saved_time) / (self.timestep*self.save_interval))
                now_n_frames = new_n_frames + old_n_frames

                if new_n_frames == 0:
                    self.lock.release()
                    return
                assert new_n_frames > 0, f"we can't write negative frames ({old_n_frames} previously in {self.checkpoint_filename}, but now only {now_n_frames})"

                frames_to_add = list()
                # there is probably a more elegant way to handle this, but this seems robust at least
                for frame in self.frames[-(new_n_frames*self.save_interval)-1:]:
                    if frame.time <= last_saved_time:
                        continue
                    if frame.time % (self.timestep * self.save_interval) == 0:
                        frames_to_add.append(frame)
                assert new_n_frames == len(frames_to_add), "pernicious math error in frame numbers!"

                new_times = np.asarray([frame.time for frame in frames_to_add])
                all_times.resize((now_n_frames,))
                all_times[-new_n_frames:] = new_times

                all_energies = h5.get("all_energies")
                new_energies = np.asarray([frame.energy for frame in frames_to_add])
                all_energies.resize((now_n_frames,))
                all_energies[-new_n_frames:] = new_energies

                new_positions = np.stack([frame.positions for frame in frames_to_add])
                all_positions = h5.get("all_positions")
                all_positions.resize((now_n_frames,n_atoms,3))
                all_positions[-new_n_frames:] = new_positions

                new_velocities= np.stack([frame.velocities for frame in frames_to_add])
                all_velocities = h5.get("all_velocities")
                all_velocities.resize((now_n_frames,n_atoms,3))
                all_velocities[-new_n_frames:] = new_velocities

                new_accels = np.stack([frame.accelerations for frame in frames_to_add])
                all_accels = h5.get("all_accelerations")
                all_accels.resize((now_n_frames,n_atoms,3))
                all_accels[-new_n_frames:] = new_accels

                new_temps = np.stack([frame.bath_temperature for frame in frames_to_add])
                all_temps = h5.get("bath_temperatures")
                all_temps.resize((now_n_frames,))
                all_temps[-new_n_frames:] = new_temps

            logger.info(f"Saving to existing checkpoint file {self.checkpoint_filename} ({new_n_frames} frames added; {last_run_time:.1f}/{self.stop_time:.1f} fs run in total)")
        else:
            with h5py.File(self.checkpoint_filename, "w") as h5:
                h5.attrs['atomic_numbers'] = self.atomic_numbers.view(np.ndarray)
                h5.attrs['masses'] = self.masses.view(np.ndarray)
                h5.attrs['finished'] = self.finished
                h5.attrs['forwards'] = self.forwards

                n_atoms = len(self.atomic_numbers)

                frames_to_add = list()
                for frame in self.frames:
                    if frame.time % (self.timestep * self.save_interval) == 0:
                        frames_to_add.append(frame)

                energies = np.asarray([frame.energy for frame in frames_to_add])
                h5.create_dataset("all_energies", data=energies, maxshape=(None,), compression="gzip", compression_opts=9)

                times = np.asarray([frame.time for frame in frames_to_add])
                h5.create_dataset("all_times", data=times, maxshape=(None,), compression="gzip", compression_opts=9)

                all_positions = np.stack([frame.positions for frame in frames_to_add])
                h5.create_dataset("all_positions", data=all_positions, maxshape=(None,n_atoms,3), compression="gzip", compression_opts=9)

                all_velocities = np.stack([frame.velocities for frame in frames_to_add])
                h5.create_dataset("all_velocities", data=all_velocities, maxshape=(None,n_atoms,3), compression="gzip", compression_opts=9)

                all_accels= np.stack([frame.accelerations for frame in frames_to_add])
                h5.create_dataset("all_accelerations", data=all_accels, maxshape=(None,n_atoms,3), compression="gzip", compression_opts=9)

                temps = np.asarray([frame.bath_temperature for frame in frames_to_add])
                h5.create_dataset("bath_temperatures", data=temps, maxshape=(None,), compression="gzip", compression_opts=9)

        logger.info(f"Saving to new checkpoint file {self.checkpoint_filename} ({len(frames_to_add)} frames added; {last_run_time:.1f}/{self.stop_time:.1f} fs run in total)")
        self.lock.release()

        # lower memory usage
        if keep_all:
            pass
        else:
            self.frames = [self.frames[-1]]

    def write_movie(self, filename, solvents="all", idxs=None):
        """
        Write a movie to a trajectory file. Detects trajectory type automatically from file extension.

        Supported file formats: ``.pdb``, ``.mol2``, ``.xyz``/``.molden``

        Args:
            filename (str): path where movie will be written
            solvents (int): number of solvent molecules to include (closest included first). can also be ``none`` or ``all``.
            idxs (list of int): indices of atoms to include. will override ``solvents`` if present.
        """

        # what do we make a movie of?
        if idxs is not None:
            assert isinstance(idxs, list), "idxs must be list of indices or ``None``!"
        else:
            if isinstance(solvents, str):
                if solvents == "high":
                    idxs = self.high_atoms
                elif solvents == "all":
                    idxs = None
                else:
                    raise ValueError(f"unknown solvents keyword {solvents} -- must be 'high' or 'all'")
            elif isinstance(solvents, int):
                molecule = self.frames[0].molecule().assign_connectivity()
                idxs = molecule.limit_solvent_shell(num_solvents=solvents, return_idxs=True)
            else:
                raise ValueError("``solvents`` must be int, 'high', or 'all'!")

        ensemble = self.as_ensemble(idxs)
        logger.info("Writing trajectory to {filename}")
        if re.search("pdb$", filename):
            cctk.PDBFile.write_ensemble_to_trajectory(filename, ensemble)
        elif re.search("mol2$", filename):
            #### connectivity matters
            ensemble.assign_connectivity()
            cctk.MOL2File.write_ensemble_to_file(filename, ensemble)
        elif re.search("molden$", filename) or re.search("xyz$", filename):
            cctk.XYZFile.write_ensemble_to_file(filename, ensemble)
        else:
            raise ValueError(f"error writing {filename}: this filetype isn't currently supported!")

    def as_ensemble(self, idxs=None):
        ensemble = cctk.ConformationalEnsemble()
        for frame in self.frames[:-1]:
            ensemble.add_molecule(frame.molecule(idxs), {"bath_temperature": frame.bath_temperature, "energy": frame.energy})
        return ensemble

    @classmethod
    def new_from_checkpoint(cls, checkpoint, frame):
        """
        Creates new trajectory from the given checkpoint file.

        Args:
            checkpoint (str): path to checkpoint file
            frame (int): the index of the desired frame

        Returns:
            new ``Trajectory`` object
        """
        assert isinstance(frame, int), "need an integer frame number"

        new_traj = cls(checkpoint_filename=checkpoint)
        new_traj.load_from_checkpoint(idxs=frame)

        assert len(new_traj.frames) == 1, "got too many frames!"
        return new_traj

    def initialize_lock(self):
        """
        Create hidden lockfile to accompany ``.chk`` file.
        """
        if self.checkpoint_filename is None:
            return

        if self.lock is None:
            lockfile = None
            if "/" in self.checkpoint_filename:
                lockfile = f"{self.checkpoint_filename}.lock"[::-1].replace("/", "./", 1)[::-1]
            else:
                lockfile = f".{self.checkpoint_filename}.lock"
            self.lock = fasteners.InterProcessLock(lockfile)

    def last_time_run(self):
        """ Get last finished time """
        return self.frames[-1].time

class ReactionTrajectory(Trajectory):
    """
    Attributes:
        termination_function (function): detects if first or last Frame has reached product/SM or should otherwise be halted.
            takes ``Frame`` argument as option and returns ``True``/``False``.
            optionally, can return 1 for forward and 2 for reverse (to differentiate recrossing from productive).
        time_after_finished (float/int): how long (in fs) to continue propagation after termination conditions reached
    """

    @classmethod
    def new_from_checkpoint(self):
        pass

    def __init__(self, termination_function=None, time_after_finished=100, **kwargs):
        super().__init__(**kwargs)

        assert isinstance(time_after_finished, (int, float)), "time_after_finished must be numeric"
        self.time_after_finished = time_after_finished
        self.elapsed_since_finished = 0

        assert hasattr(termination_function, "__call__"), "termination_function must be a function!"
        self.termination_function = termination_function

    def __str__(self):
        return f"ReactionTrajectory({len(self.frames)} frames)"

    def __repr__(self):
        return f"ReactionTrajectory({len(self.frames)} frames)"

    def initialize(self, frame=None, positions=None, velocities=None, accelerations=None, bath_temp=None, new_velocities=None, **kwargs):
        """
        Generates initial frame object for reaction trajectory. Initializes any non-zero velocities.
        Velocities are taken from the Maxwell–Boltzmann distribution for the given temperature.

        Can pass either a frame object, or the relevant attributes (frame gets precedence).

        Args:
            frame (presto.frame.Frame): equilibrated frame
            new_velocities (cctk.OneIndexedArray): array of velocities to add to equilibrated frame (for previously-frozen atoms)
            velocities, accelerations, positions (cctk.OneIndexedArray): values from frame (instead of frame object)
            bath_temperature (float):

        Returns:
            frame
        """
        logger.info("Initializing new reaction trajectory...")
        if self.has_checkpoint():
            self.load_from_checkpoint(slice(-1,None,None))
            return

        if frame is not None:
            assert isinstance(frame, presto.frame.Frame), "need a valid frame"

            positions = frame.positions
            velocities = frame.velocities
            accelerations = frame.accelerations
            bath_temp = frame.bath_temperature

        else:
            assert positions is not None, "no Frame supplied, need positions"
            assert velocities is not None, "no Frame supplied, need velocities"
            assert accelerations is not None, "no Frame supplied, need accelerations"
            assert bath_temp is not None, "no Frame supplied, need bath temperature"

            assert isinstance(positions, cctk.OneIndexedArray)
            assert isinstance(velocities, cctk.OneIndexedArray)
            assert isinstance(accelerations, cctk.OneIndexedArray)
            assert isinstance(bath_temp, (float, int, np.integer))

        new_frame = presto.frame.Frame(self, positions, velocities, accelerations, bath_temperature=bath_temp, time=0.0)

        if new_velocities is None:
            random_gaussian = np.random.normal(size=positions.shape).view(cctk.OneIndexedArray)
            random_gaussian[new_frame.active_mask()] = 0
            new_velocities = random_gaussian * np.sqrt(bath_temp * presto.constants.BOLTZMANN_CONSTANT / self.masses.reshape(-1,1))

        new_frame.velocities += new_velocities.view(cctk.OneIndexedArray)

        self.frames = [new_frame]
        self.save()

class EquilibrationTrajectory(Trajectory):
    """
    Attributes:
        bath_scheduler (function): takes current time and returns target temperature
            alternatively, pass a number if you want a constant-temperature bath
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        bath_scheduler = kwargs.get("bath_scheduler")

        if hasattr(bath_scheduler, "__call__"):
            self.bath_scheduler = bath_scheduler
        elif isinstance(bath_scheduler, (int, float)):
            def sched(time):
                return bath_scheduler

            self.bath_scheduler = sched
        else:
            raise ValueError(f"unknown type {type(bath_scheduler)} for bath_scheduler - want either a function or a number!")

    def __str__(self):
        return f"EquilibrationTrajectory({len(self.frames)} frames)"

    def __repr__(self):
        return f"EquilibrationTrajectory({len(self.frames)} frames)"

    def initialize(self, positions, velocities=None, accelerations=None, **kwargs):
        """
        Generates initial frame object for initialization trajectory.
        Velocities are taken from the Maxwell–Boltzmann distribution for the given temperature.

        Args:
            positions (cctk.OneIndexedArray): starting positions
            velocities (cctk.OneIndexedArray): starting velocities, optional.
            accelerations (cctk.OneIndexedArray): starting accelerations, optional.

        Returns:
            frame
        """
        logger.info("Initializing new equilibration trajectory...")
        if self.has_checkpoint():
            self.load_from_checkpoint(slice(-1,None,None))
            assert len(self.frames) == 1, "didn't load frames properly!"
            return

        assert isinstance(positions, cctk.OneIndexedArray), "positions must be a one-indexed array!"

        # determine active atoms
        inactive_mask = np.ones(shape=len(positions)).view(cctk.OneIndexedArray)
        inactive_mask[self.active_atoms] = 0
        inactive_mask = inactive_mask.astype(bool)

        if velocities is None:
            # add random velocity to everything
            sigma = np.sqrt(self.bath_scheduler(0) * presto.constants.BOLTZMANN_CONSTANT / self.masses.reshape(-1,1))
            velocities = np.random.normal(scale=sigma, size=positions.shape).view(cctk.OneIndexedArray)
            velocities[inactive_mask] = 0

            # subtract out center-of-mass translational motion
            com_translation = np.sum(self.masses.reshape(-1,1) * velocities, axis=0)
            correction_tran = np.tile(com_translation / np.sum(self.masses[self.active_atoms]), (len(velocities),1))
            correction_tran[inactive_mask] = 0
            velocities = velocities - correction_tran
            com_translation = np.sum(self.masses.reshape(-1,1) * velocities, axis=0)
            assert np.linalg.norm(np.sum(self.masses.reshape(-1,1) * velocities, axis=0)) < 0.0001, "didn't remove COM translation well enough!"

            velocities = velocities.view(cctk.OneIndexedArray)
        else:
            assert isinstance(velocities, cctk.OneIndexedArray)
            velocities[inactive_mask] = 0

        if accelerations is None:
            accelerations = np.zeros_like(positions).view(cctk.OneIndexedArray)
        else:
            assert isinstance(accelerations, cctk.OneIndexedArray)
            accelerations[inactive_mask] = 0

        self.frames = [presto.frame.Frame(self, positions, velocities, accelerations, bath_temperature=self.bath_scheduler(0), time=0.0)]
        self.save()

def join(traj1, traj2):
    """
    Join two reaction trajectories together -- one forward and one reverse!
    Returns a single reaction trajectory.

    Args:
        traj1 (presto.ReactionTrajectory): forward trajectory
        traj2 (presto.ReactionTrajectory): reverse trajectory

    Returns:
        combined ``ReactionTrajectory``
    """
    logger.info("Joining forward and reverse reaction trajectories....")
    assert isinstance(traj1, ReactionTrajectory), "Need a ReactionTrajectory"
    assert isinstance(traj2, ReactionTrajectory), "Need a ReactionTrajectory"

    assert traj1.forwards == True, "First trajectory must be forwards!"
    assert traj2.forwards == False, "Second trajectory must be reverse!"

    assert traj1.finished, "First trajectory must be finished!"
    assert traj2.finished, "Second trajectory must be finished!"

    assert np.array_equal(traj1.frames[0].positions, traj2.frames[0].positions), "Link positions must be same!"
    assert np.array_equal(traj1.frames[0].velocities, traj2.frames[0].velocities), "Link velocities must be same!"
    assert np.array_equal(traj1.frames[0].accelerations, traj2.frames[0].accelerations), "Link accelerations must be same!"

    new_traj = ReactionTrajectory(
        timestep = traj1.timestep,
        atomic_numbers = traj1.atomic_numbers,
        high_atoms = traj1.high_atoms,
        active_atoms = traj1.active_atoms,
        calculator = traj1.calculator,
        integrator = traj1.integrator,
        termination_function = traj1.termination_function,
        stop_time = traj1.stop_time,
        forwards = True,
    )

    f_frames = copy.deepcopy(traj1.frames)
    r_frames = copy.deepcopy(traj2.frames)
    r_frames.reverse()

    new_traj.frames = r_frames + f_frames[1:] # don't double-count middle frame

    if traj1.finished == 2:
        #### if the first traj finished with the reverse condition, reverse order
        new_traj.frames = new_traj.frames[::-1]

    return new_traj




import unittest, cctk, os
import numpy as np

import sys
sys.path.append('../presto')
import presto

if __name__ == '__main__':
    unittest.main()

class TestCheck(unittest.TestCase):
    def gen_test_traj(self):
        traj = presto.trajectory.Trajectory(
            timestep=0.5,
            atomic_numbers=cctk.OneIndexedArray([1,1]),
            high_atoms=np.array([]),
            active_atoms=np.array([1,2]),
            calculator=presto.calculators.XTBCalculator(),
            integrator=presto.integrators.VelocityVerletIntegrator(potential=presto.potentials.SphericalHarmonicPotential(radius=10)),
            stop_time=10000,
        )

        p1 = cctk.OneIndexedArray([[12,0,0],[11.5,0,0]])
        p2 = cctk.OneIndexedArray([[12,0,0],[10,0,0]])
        v = cctk.OneIndexedArray([[2,0,0],[0,0,0]])
        a = cctk.OneIndexedArray([[0,0,0],[0,0,0]])
        frame1 = presto.frame.Frame(traj, p1, v, a)
        frame2 = presto.frame.Frame(traj, p2, v, a)
        traj.frames = [frame1, frame2]
        self.assertTrue(isinstance(traj, presto.trajectory.Trajectory))
        self.assertTrue(isinstance(frame1, presto.frame.Frame))
        self.assertTrue(isinstance(frame2, presto.frame.Frame))
        return traj

    def test_radius_check(self):
        check = presto.checks.RadiusCheck()
        traj = self.gen_test_traj()
        with self.assertRaises(AssertionError):
            check.check(traj)

    def test_velocity_check(self):
        check = presto.checks.VelocityCheck()
        traj = self.gen_test_traj()
        with self.assertRaises(AssertionError):
            check.check(traj)

    def test_topology_check(self):
        check = presto.checks.TopologyCheck(interval=1)
        traj = self.gen_test_traj()
        with self.assertRaises(AssertionError):
            check.check(traj)



import unittest, cctk, os, sys
import numpy as np

sys.path.append('../presto')

import presto

if __name__ == '__main__':
    unittest.main()

class TestConstraint(unittest.TestCase):
    def gen_test_frame(self, distance):
        if os.path.exists("test/static/constraint-test.chk"):
            os.remove("test/static/constraint-test.chk")

        zs = cctk.OneIndexedArray([1, 1])
        c = presto.constraints.PairwisePolynomialConstraint(1, 2, distance, power=2)

        traj = presto.trajectory.EquilibrationTrajectory(
            timestep=0.5,
            atomic_numbers=zs,
            high_atoms=np.array([]),
            active_atoms=np.array([1, 2]),
            calculator=presto.calculators.NullCalculator([c]),
            integrator=presto.integrators.VelocityVerletIntegrator(),
            stop_time=250,
            checkpoint_filename="test/static/constraint-test.chk",
            bath_scheduler=298,
        )

        if os.path.exists("test/static/constraint-test.chk.lock"):
            os.remove("test/static/constraint-test.chk.lock")

        self.assertTrue(isinstance(traj, presto.trajectory.Trajectory))

        z = cctk.OneIndexedArray([[0,0,0], [0,0,0]])
        p = cctk.OneIndexedArray([[1,0,0], [0,0,0]])
        frame = presto.frame.Frame(traj, p, z, z)
        self.assertTrue(isinstance(frame, presto.frame.Frame))

        traj.frames = [frame]
        traj.save()
        return frame

    def test_constraint(self):
        f1 = self.gen_test_frame(1)

        f2 = f1.next()
        f3 = f2.next()

        for x, y in zip(np.ravel(f3.positions), np.ravel(f2.positions)):
            self.assertTrue(x - y == 0)

        for x, y in zip(np.ravel(f3.accelerations), np.ravel(f2.accelerations)):
            self.assertTrue(x - y == 0)

        f1 = self.gen_test_frame(1.5)
        t = f1.trajectory

        t.run(keep_all=True)

        distances = list()
        for f in t.frames:
            m = f.molecule()
            distances.append(m.get_distance(1,2))

#        os.remove(t.checkpoint_filename)

    def test_anchor(self):
        anchor = presto.constraints.Anchor(1)
        position = cctk.OneIndexedArray([[1,0,0], [0,0,0]])
        position[1] = np.asarray([1, 1, 1])

        f, e = anchor.evaluate(position)
        for x in list(f[1]):
            self.assertTrue(x + 0.0004184 < 0.000001)
        self.assertTrue(e - 0.0031379999999999997 < 0.000001)



import numpy as np
import math, copy, cctk

import presto

class Constraint():

    def __init__(self):
        pass

    def evaluate(self, positions):
        """
        Returns forces and energy.
        """
        pass

class PairwisePolynomialConstraint(Constraint):
    """
    Represents a pairwise constraint of the form: – energy = ``constant`` / ``power`` * (distance - ``equilibrium``) ** ``power``.

    This will lead to a force of the form ``constant`` * (distance - ``equilibrium``) ** (``power`` - 1).

    If ``atom1`` or ``atom2`` is a list, then the closest/farthest distance will be taken. This corresponds to Singleton's "interlocking sphere biasing potential".

    Attributes:
        atom1 (int): number of 1st atom
        atom2 (int): number of 2nd atom
        power (float): dependence on distance
        force_constant (float): force constant
        equilibrium (float): desired distance in Å
        min (bool): whether the largest or smallest distance should be taken.
    """

    def __init__(self, atom1, atom2, equilibrium, power=2, force_constant=10, convert_from_kcal=True, min=True):
        assert isinstance(atom1, (list, int)), "atom number must be integer"
        assert isinstance(atom2, (list, int)), "atom number must be integer"
        assert isinstance(power, (int, float)), "power must be numeric"
        assert isinstance(force_constant, (int, float)), "force_constant must be numeric"
        assert isinstance(equilibrium, (int, float)), "equilibrium must be numeric"

        if isinstance(atom1, list):
            assert all(isinstance(x, int) for x in atom1)
        if isinstance(atom2, list):
            assert all(isinstance(x, int) for x in atom2)

        if convert_from_kcal:
            force_constant *= 0.0004184

        self.atom1 = atom1
        self.atom2 = atom2
        self.power = power
        self.force_constant = force_constant
        self.equilibrium = equilibrium

        if min:
            self.min = True
        else:
            self.min = False

    def __str__(self):
        return f"PairwisePolynomialConstraint(atom1={self.atom1}, atom2={self.atom2}, equilibrium={self.equilibrium:.3f}, power={self.power}, force_constant={self.force_constant:.5f})"

    def __repr__(self):
        return f"PairwisePolynomialConstraint(atom1={self.atom1}, atom2={self.atom2}, equilibrium={self.equilibrium:.3f}, power={self.power}, force_constant={self.force_constant:.5f})"

    def evaluate(self, positions):
        assert isinstance(positions, cctk.OneIndexedArray), "positions must be one-indexed array"
        x1, x2 = None, None

        atoms1 = self.atom1
        atoms2 = self.atom2
        if isinstance(atoms1, int):
            atoms1 = [atoms1]
        if isinstance(atoms2, int):
            atoms2 = [atoms2]

        # this is where we find the right pair of atoms
        if self.min == True:
            min_d = 100
            for x in atoms1:
                for y in atoms2:
                    d = np.linalg.norm(positions[x] - positions[y])
                    if d < min_d:
                        min_d = d
                        x1 = positions[x]
                        x2 = positions[y]
        else:
            max_d = 0
            for x in atoms1:
                for y in atoms2:
                    d = np.linalg.norm(positions[x] - positions[y])
                    if d > max_d:
                        max_d = d
                        x1 = positions[x]
                        x2 = positions[y]

        delta = np.linalg.norm(x1-x2) - self.equilibrium
        direction = (x2 - x1)/np.linalg.norm(x1-x2)

        forces = np.zeros_like(positions.view(np.ndarray)).view(cctk.OneIndexedArray)
        forces[self.atom1] = delta ** (self.power - 1) * self.force_constant * 1 * direction
        forces[self.atom2] = delta ** (self.power - 1) * self.force_constant * -1 * direction

        energy = self.force_constant / self.power * delta ** self.power

        return forces, energy

class Anchor(Constraint):
    """
    Applies a weak potential fixing an atom towards a given coordinate in space.

    Analogous to Singleton's "zeroatom" keyword.

    Attributes:
        atom (int): number of atom
        coordinate (np.ndarray): desired coordinate
        power (float): dependence on distance
        force_constant (float): force constant
    """

    def __init__(self, atom, coordinate=np.zeros(3), power=2, force_constant=5, convert_from_kcal=True):
        assert isinstance(atom, int), "atom number must be integer"
        assert isinstance(coordinate, np.ndarray), "coordinate must be 3d np.ndarray"
        assert isinstance(power, (int, float)), "power must be numeric"
        assert isinstance(force_constant, (int, float)), "force_constant must be numeric"

        if convert_from_kcal:
            force_constant *= 0.0004184

        self.atom = atom
        self.coordinate = coordinate
        self.power = power
        self.force_constant = force_constant

    def evaluate(self, positions):
        assert isinstance(positions, cctk.OneIndexedArray), "positions must be one-indexed array"
        x = positions[self.atom]

        delta = np.linalg.norm(x)
        direction = (-1 * x)/np.linalg.norm(x)

        forces = np.zeros(positions.shape).view(cctk.OneIndexedArray)
        f = delta ** (self.power - 1) * self.force_constant * direction
        forces[self.atom] = f

        energy = self.force_constant / self.power * delta ** self.power

        return forces, energy

def build_constraints(settings):
    """
    Build constraint list from settings dict.
    """
    constraints = list()
    for row in list(settings.values()):
        assert "atom1" in row, "need ``atom1`` defined for constraint!"
        assert isinstance(row["atom1"], (int, list)), "``atom1`` must be integer or list of integers!"
        assert "atom2" in row, "need ``atom2`` defined for constraint!"
        assert isinstance(row["atom2"], (int, list)), "``atom2`` must be integer or list of integers!"
        assert "equilibrium" in row, "need equilibrium distance ``equilibrium`` defined for constraint!"
        assert isinstance(row["equilibrium"], (int, float)), "``equilibrium`` must be numeric!"

        args = {"atom1": row["atom1"], "atom2": row["atom2"], "equilibrium": row["equilibrium"]}

        if "power" in row:
            assert isinstance(row["power"], int), "``power`` must be integer!"
            args["power" ] = row["power"]

        if "force_constant" in row:
            assert isinstance(row["force_constant"], (int, float)), "``force_constant`` must be numeric"
            args["force_constant"] = row["force_constant"]

        if "which" in row:
            assert row["which"] in ["min", "max"], "`which` must be `min` or `max`"
            if row["which"] == "min":
                args["min"] = True
            else:
                args["min"] = False

        constraints.append(PairwisePolynomialConstraint(**args))
    return constraints





import presto, logging, time
import numpy as np

logger = logging.getLogger(__name__)

class Controller():
    """
    Controls actual running of frames. Dispatches to other objects.
    """
    def __init__(self, trajectory):
        assert isinstance(trajectory.calculator, presto.calculators.Calculator)
        assert isinstance(trajectory.integrator, presto.integrators.Integrator)
        assert all([isinstance(c, presto.checks.Check) for c in trajectory.checks])
        assert all([isinstance(r, presto.reporters.Reporter) for r in trajectory.reporters])

        self.trajectory = trajectory

    def run(self, checkpoint_interval=25, end_time=None, runtime=None, **kwargs):
        current_time = self.trajectory.frames[-1].time
        dt = self.trajectory.timestep
        interval = self.trajectory.save_interval

        if end_time is None:
            if runtime is not None:
                assert isinstance(runtime, (int, float)), "runtime must be numeric"
                end_time = min(current_time + runtime, self.trajectory.stop_time)
            else:
                end_time = self.trajectory.stop_time
        else:
            assert isinstance(end_time, (int, float)), "end_time must be numeric"

        assert end_time >= current_time, f"error: end_time {end_time} must be greater than or equal to current_time {current_time}"

        if end_time == current_time:
            return

        count = 0
        finished_early = False
        logger.info(f"Trajectory will run {int((end_time-current_time)/self.trajectory.timestep)} frames (current time = {current_time:.1f} fs, end time = {end_time:.1f} fs)") 
        while current_time < end_time:
            current_time += dt
            current_frame = self.trajectory.frames[-1]

            bath_temperature = current_frame.bath_temperature
            if isinstance(self.trajectory, presto.trajectory.EquilibrationTrajectory):
                bath_temperature = self.trajectory.bath_scheduler(current_time)

            new_frame = None
            try:
                new_frame = current_frame.next(forwards=self.trajectory.forwards, temp=bath_temperature)
            except Exception as e:
                logger.info(f"Error at time {current_time} - terminating run")
                raise ValueError(f"Controller failed: {e}")
            assert new_frame.time == current_time, f"frame time {new_frame.time} does not match loop time {current_time}"
            self.trajectory.frames.append(new_frame)

            for check in self.trajectory.checks:
                if int(current_time % check.interval) == 0:
                    check.check_frame(new_frame)

            for reporter in self.trajectory.reporters:
                if int(current_time % reporter.interval) == 0:
                    reporter.report(self.trajectory)

            # do we initiate early stopping?
            if not finished_early:
                if isinstance(self.trajectory, presto.trajectory.ReactionTrajectory):
                    if self.trajectory.termination_function(self.trajectory.frames[-1]):
                        end_time = current_time + self.trajectory.time_after_finished
                        finished_early = True
                        logger.info(f"Reaction trajectory finished! {self.trajectory.time_after_finished} additional fs will be run.")

            if int(current_time/dt) % int(checkpoint_interval/dt) == 0:
                self.trajectory.save()

            count += 1
            if count < 10:
                logger.info(f"Run initiated ok - frame {count:05d} completed in {new_frame.elapsed:.2f} s.")

        self.trajectory.save()
        if current_time == self.trajectory.stop_time:
            self.trajectory.finished = True
        elif finished_early:
#            self.trajectory.finished = self.trajectory.termination_function(self.trajectory.frames[-1])
            self.trajectory.finished = True # somehow that previous line was not working

        logger.info(f"Trajectory done running with {self.trajectory.num_frames()} frames.")
        return



import presto.config
import presto.constants
import presto.calculators
import presto.integrators
import presto.frame
import presto.trajectory
import presto.analysis
import presto.constraints
import presto.replica_exchange
import presto.controller
import presto.checks
import presto.reporters
import presto.potentials



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="presto-md",
    packages=["presto", "presto.solvents"],
    package_data={"presto.solvents": ["*"],},
    version="0.2.3",
    author="Corin Wagen",
    author_email="corin.wagen@gmail.com",
    license="Apache 2.0",
    description="explicit solvent molecular dynamics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cwagen/presto",
#    packages=setuptools.find_packages(),
    install_requires=["cctk", "tqdm", "h5py", "pyyaml", "matplotlib", "asciichartpy", "fasteners"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Science/Research",
    ],
    python_requires='>=3.6',
)



import numpy as np
import math, copy, cctk
from scipy import constants

import presto

class Integrator():
    def next(self, frame, forwards=True):
        pass

class VelocityVerletIntegrator(Integrator):
    """
    Defines a Velocity Verlet integrator (NVE ensemble).

    Attributes:
        potential (presto.potentials.Potential): confining potential
    """

    def __init__(self, potential=None):
        if potential is not None:
            assert isinstance(potential, presto.potentials.Potential), f"needed a presto.Potential, got a {potential} instead"
        self.potential = potential

    def next(self, frame, forwards=True):
        calculator = frame.trajectory.calculator
        timestep = frame.trajectory.timestep
        if forwards == False:
            timestep = timestep * -1

        try:
            x_full = frame.positions + frame.velocities * timestep + 0.5 * frame.accelerations * (timestep ** 2)

            energy, forces = calculator.evaluate(frame.trajectory.atomic_numbers, x_full, frame.trajectory.high_atoms)
            if self.potential is not None:
                pe, pf = self.potential.evaluate(x_full)
                forces += pf
                energy += pe
            forces[frame.inactive_mask()] = 0

            a_full = forces / frame.masses()

            v_full = frame.velocities + (frame.accelerations + a_full) * 0.5 * timestep

            return energy, x_full, v_full, a_full

        except Exception as e:
            raise ValueError(f"Error in integrator: {e}")


class LangevinIntegrator(VelocityVerletIntegrator):
    """
    Defines a Langevin integrator (NVT ensemble).

    Attributes:
        viscosity (float): solvent viscosity (in amu/(fs*Å))
        radius (float): distance from center where this kicks in!
        potential (presto.potentials.Potential): confining potential
    """

    def __init__(self, viscosity=0.001, convert_from_pascal_seconds=True, radius=0, potential=None):
        assert isinstance(viscosity, (int, float)), "viscosity must be numeric"
        if convert_from_pascal_seconds:
            viscosity = viscosity * presto.constants.AMU_A_FS_PER_PASCAL_SECOND
        self.viscosity = viscosity

        assert isinstance(radius, (int, float)), "radius must be numeric"
        self.radius = radius

        if potential is not None:
            assert isinstance(potential, presto.potentials.Potential), f"needed a presto.Potential, got a {potential} instead"
        self.potential = potential

    def next(self, frame, forwards=True):
        """
        Using the approach from http://itf.fys.kuleuven.be/~enrico/Teaching/molecular_dynamics_2015.pdf
        (Vanden-Eijden–Ciccotti second-order Langevin integrator)
        """
        calculator = frame.trajectory.calculator
        timestep = frame.trajectory.timestep
        if forwards == False:
            timestep = timestep * -1

        xi = 6 * math.pi * self.viscosity * frame.radii() / frame.trajectory.masses
        sigma = np.sqrt(2 * xi * presto.constants.BOLTZMANN_CONSTANT * frame.bath_temperature / frame.trajectory.masses)

        # exclude those beyond the radius
        no_apply_to = np.linalg.norm(frame.positions, axis=1) < self.radius
        xi[no_apply_to] = 0
        sigma[no_apply_to] = 0
        xi = xi.reshape(-1,1)
        sigma = sigma.reshape(-1,1)

        rand1 = np.random.normal(size=frame.positions.shape).view(cctk.OneIndexedArray)
        rand2 = np.random.normal(size=frame.positions.shape).view(cctk.OneIndexedArray)

        C = 0.5 * (timestep ** 2) * (frame.accelerations - (xi * frame.velocities)) + sigma * (timestep ** 1.5) * ((0.5 * rand1) + (rand2 / (2 * math.sqrt(3))))

        x_full = frame.positions + timestep * frame.velocities + C
        x_full[frame.inactive_mask()] = frame.positions[frame.inactive_mask()] # stay still!

        energy, forces = calculator.evaluate(frame.trajectory.atomic_numbers, x_full, frame.trajectory.high_atoms)
        if self.potential is not None:
            pe, pf = self.potential.evaluate(x_full)
            forces += pf
            energy += pe

        forces[frame.inactive_mask()] = 0
        a_full = forces / frame.masses()

        v_full = frame.velocities + 0.5 * timestep * (a_full + frame.accelerations) - timestep * xi * frame.velocities + sigma * math.sqrt(timestep) * rand1 - xi * C
        v_full[frame.inactive_mask()] = 0

        return energy, x_full, v_full, a_full

def build_integrator(settings, potential=None):
    """
    Build integrator from settings dict.
    """
    assert isinstance(settings, dict), "Need to pass a dictionary!!"
    assert "type" in settings, "Need `type` for integrator"
    assert isinstance(settings["type"], str), "Integrator `type` must be a string"

    if settings["type"].lower() == "verlet":
        return VelocityVerletIntegrator(potential=potential)
    elif settings["type"].lower() == "langevin":
        assert "viscosity" in settings, "Need `viscosity` for Langevin integrator!"
        assert isinstance(settings["viscosity"], (int, float)), "Integrator `viscosity` must be numeric."

        radius = 0
        if "radius" in settings:
            assert isinstance(settings["radius"], (int, float)), "Integrator `radius` must be numeric."
            radius = settings["radius"]

        return LangevinIntegrator(settings["viscosity"], radius=radius, potential=potential)
    else:
        raise ValueError(f"Unknown integrator type {settings['type']}! Allowed options are `verlet` or `langevin`.")




import unittest, cctk
import numpy as np

import sys
sys.path.append('../presto')

import presto

if __name__ == '__main__':
    unittest.main()

class TestLangevinIntegrator(unittest.TestCase):
    def gen_test_frame(self):
        zs = cctk.OneIndexedArray([2])
        traj = presto.trajectory.Trajectory(
            timestep=0.5,
            atomic_numbers=zs,
            high_atoms=np.array([]),
            active_atoms=np.array([1]),
            stop_time=100,
            calculator=presto.calculators.XTBCalculator(),
            integrator=presto.integrators.LangevinIntegrator(viscosity=0.0001),
        )

        self.assertTrue(isinstance(traj, presto.trajectory.Trajectory))

        e0 = cctk.OneIndexedArray([[0,0,0]])
        e1 = cctk.OneIndexedArray([[1,0,0]])
        frame = presto.frame.Frame(traj, e0, e1, e0)
        self.assertTrue(isinstance(frame, presto.frame.Frame))
        return frame

    def test_radius(self):
        frame = self.gen_test_frame()
        self.assertEqual(frame.radii()[1], 1.4)

#    def test_drag(self):
#        frame = self.gen_test_frame()
#        new_a = cctk.OneIndexedArray([1,0,0])
#        print(frame.trajectory.integrator.drag_forces(frame, True, new_a))

class TestSphericalIntegration(unittest.TestCase):
    def test_spherical_harmonic_potential(self):
        potential = presto.potentials.SphericalHarmonicPotential(radius=10, force_constant=1, convert_from_kcal=False)

        x = np.array([[1], [9], [11], [-12]])
        e, f = potential.evaluate(x)

        self.assertEqual(f[1], 0)
        self.assertEqual(f[2], 0)
        self.assertEqual(f[3], -1)
        self.assertEqual(f[4], 2)
        self.assertEqual(e,2.5)

        x = np.array([[0,0,10]])
        e, f = potential.evaluate(x)
        self.assertEqual(np.linalg.norm(f[1]), 0)
        self.assertEqual(e,0)

        x = np.array([[0,0,-8]])
        e, f = potential.evaluate(x)
        self.assertEqual(np.linalg.norm(f[1]), 0)
        self.assertEqual(e,0)

        x = np.array([[0,0,1]])
        e, f = potential.evaluate(x)
        self.assertEqual(np.linalg.norm(f[1]), 0)
        self.assertEqual(e,0)



import numpy as np
import math, copy, cctk, time

import cctk
from cctk.helper_functions import get_symbol, get_vdw_radius

import presto

class Frame():
    """
    Represents one frame in a trajectory.

    Attributes:
        trajectory (presto.Trajectory):
        positions (cctk.OneIndexedArray):
        velocities (cctk.OneIndexedArray):
        accelerations (cctk.OneIndexedArray):

        time (float): time of current frame within trajectory, in fs
        energy (float):
        bath_temperature (float):

        elapsed (float): wallclock time to run frame
    """

    def __init__(self, trajectory, x, v, a, time=0, bath_temperature=298, energy=0.0, elapsed=0):
        assert isinstance(trajectory, presto.trajectory.Trajectory), "need trajectory"

        assert len(x) == len(v), "length of positions not same as length of velocities!"
        assert len(x) == len(a), "length of positions not same as length of accelerations!"

        assert isinstance(x, cctk.OneIndexedArray), "positions is not a one-indexed array!"
        assert isinstance(v, cctk.OneIndexedArray), "velocities is not a one-indexed array!"
        assert isinstance(a, cctk.OneIndexedArray), "accelerations is not a one-indexed array!"

        assert (x.ndim == 2) and (x.shape[1] ==  3), "positions must be an n x 3 ndarray"
        assert (v.ndim == 2) and (v.shape[1] ==  3), "velocities must be an n x 3 ndarray"
        assert (a.ndim == 2) and (a.shape[1] ==  3), "accelerations must be an n x 3 ndarray"

        assert isinstance(bath_temperature, (float, int, np.integer)), "bath temperature needs to be numeric!"
        assert bath_temperature >= 0, "bath temperature must be positive or 0"

        assert isinstance(elapsed, (float, int, np.integer)), "elapsed needs to be numeric!"

        if time is not None:
            assert isinstance(time, (float, int, np.integer)), "time must be numeric"
            time = float(time)

        self.trajectory = trajectory
        self.positions = x
        self.velocities = v
        self.accelerations = a
        self.bath_temperature = bath_temperature
        self.energy = energy
        self.time = time
        self.elapsed = elapsed

    def __str__(self):
        return f"Frame({len(self.positions)} atoms, time={self.time:.1f})"

    def __repr__(self):
        return f"presto.frame.Frame({len(self.positions)} atoms, time={self.time:.1f})"

    def next(self, temp=None, forwards=True):
        """
        Computes next frame using ``self.trajectory.integrator``.

        The desired bath temperature is not used in the current force calculations, but is passed to the output frame.
        """
        if temp is None:
            temp = self.bath_temperature
        assert isinstance(temp, (float, int, np.integer)), "temp must be numeric!"

        try:
            start = time.time()
            energy, new_x, new_v, new_a = self.trajectory.integrator.next(self, forwards=forwards)
            end = time.time()
            elapsed = end - start

            # strictly speaking the energy is for this frame, but we'll give the next frame this energy too in case it's the last one (better than leaving it null).
            self.energy = energy
            return Frame(self.trajectory, new_x, new_v, new_a, bath_temperature=temp, time=self.time+self.trajectory.timestep, energy=energy, elapsed=elapsed)
        except Exception as e:
            raise ValueError(f"Error in frame.next(): {e}")

    def prev(self, temp=None):
        """
        Computes previous frame using ``self.trajectory.integrator``.

        The desired bath temperature is not used in the current force calculations, but is passed to the output frame.
        """
        return self.next(temp, forwards=False)

    def potential_energy(self):
        """ Returns the potential energy in kcal/mol. """
        return self.energy * presto.constants.KCAL_PER_HARTREE

    def kinetic_energy(self, convert_to_kcal=True):
        """ Returns the kinetic energy in kcal/mol. """
        v = np.linalg.norm(self.velocities[self.trajectory.active_atoms], axis=1)
        m = self.trajectory.masses.view(cctk.OneIndexedArray)[self.trajectory.active_atoms]
        K = np.multiply(m.view(np.ndarray), np.power(v, 2))
        if convert_to_kcal:
            return np.sum(K) / (2 * presto.constants.AMU_A2_FS2_PER_KCAL_MOL)
        else:
            return np.sum(K) / 2

    def total_energy(self):
        """ Returns the total energy. """
        return self.kinetic_energy() + self.potential_energy()

    def temperature(self):
        """
        Computes the instantaneous temperature based on the equipartition theorem, counting only the active atoms.

        T = sum{ m_i * v_i ** 2} / (kB * Nf)
        """
        K = self.kinetic_energy(convert_to_kcal=False) * 2
        F = 3 * len(self.trajectory.active_atoms) - 3

        if F == 0: # single particle
            F = 3

        return float(K / (F * presto.constants.BOLTZMANN_CONSTANT))

    def pressure(self):
        """
        Computes the pressure based on the following formula:

        P = 1/(3*V) * (\sum{m_i * v_i * v_i + r_i * f_i}
        """
        m = self.trajectory.masses.view(cctk.OneIndexedArray)[self.trajectory.active_atoms].view(np.ndarray)
        v = self.velocities[self.trajectory.active_atoms].view(np.ndarray)
        a = self.accelerations[self.trajectory.active_atoms].view(np.ndarray)
        x = self.positions[self.trajectory.active_atoms].view(np.ndarray)

        volume = self.volume()

        v = np.linalg.norm(self.velocities[self.trajectory.active_atoms], axis=1)
        m = self.trajectory.masses.view(cctk.OneIndexedArray)[self.trajectory.active_atoms]
        K = np.multiply(m.view(np.ndarray), np.power(v, 2))

        pressure = np.sum(K) / (3 * volume)
        for pos, mass, accel in zip(x, m, a):
            pressure += np.dot(pos, mass * accel) / (3 * volume)

        return pressure / presto.constants.AMU_A_FS2_PER_ATM

    def volume(self):
        return self.molecule().volume(qhull=True)

    def inactive_mask(self):
        """
        Returns an ``np.ndarray`` of the same length as ``positions`` where every active atom is ``False`` and every inactive atom is ``True``.
        """
        inactive_mask = np.ones(shape=len(self.positions)).view(cctk.OneIndexedArray)
        inactive_mask[self.trajectory.active_atoms] = 0
        inactive_mask  = inactive_mask.astype(bool)
        return inactive_mask

    def active_mask(self):
        """
        Returns an ``np.ndarray`` of the same length as ``positions`` where every active atom is ``False`` and every inactive atom is ``True``.
        """
        active_mask = np.zeros(shape=len(self.positions)).view(cctk.OneIndexedArray)
        active_mask[self.trajectory.active_atoms] = 1
        active_mask  = active_mask.astype(bool)
        return active_mask

    def molecule(self, idxs=None):
        if idxs is not None:
            return cctk.Molecule(self.trajectory.atomic_numbers[idxs], self.positions[idxs])
        else:
            return cctk.Molecule(self.trajectory.atomic_numbers, self.positions)

    def remove_com_motion(self):
        # move centroid to origin
        centroid = np.mean(self.positions, axis=0)
        self.positions = self.positions - centroid
        assert np.linalg.norm(np.sum(self.positions, axis=0)) < 0.0001, "didn't center well enough!"

        # subtract out center-of-mass translational motion (linear momentum)
        com_translation = np.sum(self.masses() * self.velocities, axis=0)
        correction_tran = np.tile(com_translation / np.sum(self.masses()), (len(self.velocities),1))
        self.velocities = self.velocities - correction_tran
        assert np.linalg.norm(np.sum(self.masses() * self.velocities, axis=0)) < 0.0001, "didn't remove COM translation well enough!"

        return self

    def masses(self):
        return self.trajectory.masses.reshape(-1,1).view(cctk.OneIndexedArray)

    def radii(self):
        vdw_radii = {z: get_vdw_radius(z) for z in set(self.trajectory.atomic_numbers)}
        radii_by_num = np.array([vdw_radii[z] for z in self.trajectory.atomic_numbers]).view(cctk.OneIndexedArray)
        return radii_by_num

    def L(self):
        """ Angular momentum """
        return np.sum(np.cross(self.velocities, self.masses() * self.positions), axis=0)
